from __future__ import annotations
from typing import List, Optional, Tuple
import pandas as pd
import json
from continual.python.sdk import featuresets
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.model_versions import ModelVersion, ModelVersionManager
from continual.python.sdk.featuresets import FeatureSet
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.events import EventManager

from continual.python.common.query import Query

from continual.python.sdk.batchpredictions import (
    BatchPrediction,
    BatchPredictionManager,
)


class ModelManager(Manager):
    """Manages model resources."""

    name_pattern: str = "projects/{project}/models/{model}"

    def get(self, id: str) -> Model:
        """Get model.

        Arguments:
            id: Model name or id.

        Returns
            An experiment.
        """
        req = management_pb2.GetModelRequest(name=self.name(id))
        resp = self.client._management.GetModel(req)
        return Model.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[Model]:
        """List model.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of models.
        """
        req = management_pb2.ListModelsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListModels(req)
        return [Model.from_proto(x, client=self.client) for x in resp.models]

    def list_all(self) -> Pager[Model]:
        """List all model.

        Pages through all model using an iterator.

        Returns:
            A iterator of all model.
        """

        def next_page(next_page_token):
            req = management_pb2.ListModelsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListModels(req)
            return (
                [Model.from_proto(x, client=self.client) for x in resp.models],
                resp.next_page_token,
            )

        return Pager(next_page)

    def train(self, id: str) -> ModelVersion:
        """Train a model.

        Creates a new model version and starts a training run using
        the lastest data available in the feature store.

        Arguments:
            id: Name or id of model.

        Returns:
            A new model version.
        """
        req = management_pb2.TrainModelRequest(name=self.name(id))
        resp = self.client._management.TrainModel(req)
        return ModelVersion.from_proto(resp, client=self.client)

    def delete(self, id: str):
        """Delete a model.

        Deletes the model from system.

        Arguments:
            id: Name or id of model.

        Returns:
            None
        """
        req = management_pb2.DeleteModelRequest(name=self.name(id))
        self.client._management.DeleteModel(req)
        return

    def get_related_feature_sets(self, project: str, entity: str) -> List[FeatureSet]:
        """Get get all feature sets from same entity as a feature set.

        Arugments:
            id: Name or id of feature set.
        Returns:
            A list of feature sets.
        """
        req = management_pb2.GetEntityFeatureSetsRequest(
            project_name=project, entity_id=entity
        )
        resp = self.client._management.GetEntityFeatureSets(req)
        return [FeatureSet.from_proto(x, client=self.client) for x in resp.feature_sets]

    def _get_featureset_map(self):
        return None

    def get_training_data(
        self,
        model,
        n: Optional[int] = None,
        normalized: bool = False,
        metadata: bool = False,
        latest: bool = False,
        all: bool = False,
    ) -> Tuple[pd.DataFrame, Query]:
        """Get training data for a model.

        Arguments:
            id: Name or id of model.
            n: Number of records.
            normalize: Only return normalized data.
            metadata: Include metadata.
            latest: Only return latest data for each entity.
        Returns:
            Training DataFrame and serving query tuple.
        """
        proj = self.client.projects.get(self.parent)
        all_fs = proj.feature_sets.list_all()
        featureset_map = {}
        for f in all_fs:
            featureset_map[f.name] = f.to_proto()

        query = Query.form_query(
            model=model.to_proto(),
            metadata=metadata,
            all_featuresets=featureset_map,
        )

        if n == 0:
            return pd.DataFrame(), query

        # TODO: add paging and more options around stard/end times

        req = management_pb2.GetTrainingDataRequest(
            query=query.to_proto(),
        )

        resp = self.client._management.GetTrainingData(req)

        e_dict = []
        if resp.value is not None:
            for k in resp.value:
                feature_dict = json.loads(k)
                # Reconcile response to fill missing columns
                self._serving_data_reconcile(query, feature_dict)
                e_dict.append(feature_dict)

        return (pd.DataFrame(e_dict), query)

    def get_serving_data(
        self,
        query: Query,
        overrides: Optional[dict] = None,
    ) -> dict:
        """Get online serving data.

        Arguments:
            query: Serving data query.
            id: Entity id.
            overrides: Local data overrides.

        Returns:
            A dictionary of online features.
        """
        q = query.clone()

        # TODO: this should pass in typed overrides.
        # when we have a grpc predict call we can do this more easily
        overide_str = None
        if overrides is not None:
            overide_str = json.dumps(overrides)
        req = management_pb2.GetServingDataRequest(
            query=q.to_proto(), overrides=overide_str
        )

        serving_data = self.client._management.GetServingData(req)
        # TODO: factor out from training/serving code.
        e_dict = []
        if serving_data.value is not None:
            for k in serving_data.value:
                feature_dict = json.loads(k)
                # Reconcile response to fill missing columns
                self._serving_data_reconcile(q, feature_dict)
                e_dict.append(feature_dict)
        return e_dict

    def _serving_data_reconcile(self, query, e_dict):
        for field in query.fields:
            if field.qualified_name not in e_dict:
                e_dict[field.qualified_name] = None

            if field.relation is not None:
                self._serving_data_reconcile(field.relation, e_dict)

        if query.related is None:
            return

        for rel in query.related:
            self._serving_data_reconcile(rel, e_dict)

        return

    def batch_predict(
        self,
        model_id: str,
        source_type: str,
        source: str,
        dest_type: str,
        incremental: bool = False,
        prediction_model: str = None,
    ) -> BatchPrediction:
        """Deploy model version.

        Arguments:
            model_id: Name or id of model.
            source_type:
            source:
            dest_type:
        Returns:
            A batch prediction job object to track batch prediction status.
        """
        model_name = self.name(model_id)
        req = management_pb2.CreateBatchPredictionRequest(
            model=model_name,
            source_type=source_type,
            source=source,
            dest_type=dest_type,
            incremental=incremental,
            prediction_model=prediction_model,
        )
        resp = self.client._management.BatchPredict(req)
        return BatchPrediction.from_proto(resp, client=self.client)

    def get_data(
        self, model: Model, page_size: int = None, page_token: str = None
    ) -> Tuple[pd.DataFrame, str]:
        """Get data for the featureset

        Arguments:
            page_size: number of rows to fetch.
            page_token: starting token for the next page of data.

        Returns:
            A dictionary of features.
        """
        req = management_pb2.GetDataRequest(
            name=model.name,
            page_size=page_size,
            page_token=page_token,
        )

        resp = self.client._management.GetData(req)
        # print("resp: {}".format(resp))
        df = None
        rows = []
        if resp.value is not None:
            for k in resp.value:
                feature_dict = json.loads(k)
                rows.append(feature_dict)

        if len(rows) > 0:
            df = pd.DataFrame(rows)

        return df, resp.next_page_token

    def diff(
        self,
        right_yaml: str,
        right_project: str,
        right_environment: str,
        right_model: str,
        left_yaml: str,
        left_project: str,
        left_environment: str,
        left_model: str,
    ):
        """Get data for the featureset

        Arguments:
            right_yaml
            right_model
            right_feature_set
            left_yaml
            left_project
            left_model

        Returns:
            Diff Report
        """
        return featuresets.diff(
            right_yaml,
            right_project,
            right_environment,
            right_model,
            left_yaml,
            left_project,
            left_environment,
            left_model,
            models=True,
        )

    def get_signature(self, id):
        req = management_pb2.GetModelQueryRequest(name=self.name(id))
        resp = self.client._management.GetModelQuery(req)
        return resp.query_json


class Model(Resource, types.Model):
    """Model resource."""

    name_pattern: str = "projects/{project}/models/{model}"
    manager: ModelManager

    model_versions: ModelVersionManager
    """Model version manager."""

    promotions: PromotionManager
    """Promotion manager."""

    batch_prediction_jobs: BatchPredictionManager
    """Batch Prediction  manager."""

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = ModelManager(parent=self.parent, client=self.client)
        self.model_versions = ModelVersionManager(parent=self.name, client=self.client)
        self.promotions = PromotionManager(parent=self.name, client=self.client)
        self.batch_prediction_jobs = BatchPredictionManager(
            parent=self.name, client=self.client
        )
        self.events = EventManager(client=self.client, parent=self.name)

    def get_training_data(
        self,
        n: Optional[int] = None,
        normalized: bool = False,
        metadata: bool = False,
        latest: bool = False,
        all: bool = False,
    ) -> Tuple[pd.DataFrame, Query]:
        """Get training data for model.

        Arguments:
            n: Number of records.
            normalized: Return only base data.
            metadata: Include metadata.
        Returns:
            A training data set.
        """
        return self.manager.get_training_data(
            self,
            n=n,
            normalized=normalized,
            metadata=metadata,
            latest=latest,
            all=all,
        )

    def get_serving_data(self, query: Query, overrides: Optional[dict] = None):
        """Get online serving data.

        Arguments:
            query: Serving query token.
            overrides: Local data overrides
        Returns:
            DataFrame of live data.
        """
        return self.manager.get_serving_data(query, overrides)

    def train(self) -> ModelVersion:
        """Wait for experiment to complete."""
        return self.manager.train(self.id)

    def delete(self):
        """Wait for experiment to complete."""
        return self.manager.delete(self.id)

    def batch_predict(
        self,
        source_type: str = "STORE",
        source: str = "",
        dest_type: str = "STORE",
        incremental: bool = False,
        prediction_model: str = None,
    ) -> BatchPrediction:
        """Start batch prediction

        Arguments:
            source_type:
            source: uri of file with records to be batch predicted
            dest_type: proto enum for destination type type
            dest_type:

        Returns:
            A batch prediction job object to track batch prediction status.
        """
        return self.manager.batch_predict(
            self.id,
            source_type,
            source,
            dest_type,
            incremental=incremental,
            prediction_model=prediction_model,
        )

    @property
    def feature_set(self) -> FeatureSet:
        """Expanded feature set for model."""
        return self.client.feature_sets.get(super().feature_set)

    @feature_set.setter
    def feature_set(self, value: FeatureSet) -> None:
        types.Model.feature_set.fset(self, value)

    def get_data(self, page_size=None, page_token=None):
        return self.manager.get_data(self, page_size, page_token)

    def get_signature(self):
        return self.manager.get_signature(self.id)



import typer
import logging
import subprocess
import os
import yaml
import pandas as pd

from continual.python.sdk.exceptions import BaseException
from continual.python.sdk.identifiers import ProjectEnvironmentIdentifer
from continual.rpc.management.v1 import types
from continual import Client

from enum import Enum
from functools import wraps
from git import Repo
from cron_descriptor import get_description

from rich.table import Table
from rich.console import Console
from rich.style import Style
from rich.layout import Layout

logger = logging.getLogger("continual.cli.utils")

continual_style_green = Style(color="rgb(157,209,186)")
continual_style_blue = Style(color="rgb(96,238,252)")
continual_style_purple = Style(color="rgb(63,47,159)")
continual_style_dark_blue = Style(color="rgb(5,27,71)")
continual_style_black = Style(color="rgb(0,0,0)")
continual_style_white = Style(color="rgb(255,255,255)")


class ContinualStyle(Enum):
    WHITE = "WHITE"
    BLACK = "BLACK"
    BLUE = "BLUE"
    GREEN = "GREEN"
    PURPLE = "PURPLE"
    DARK_BLUE = "DARK_BLUE"


class PerformanceMetric(str, Enum):
    RMSE = "rmse"
    MAE = "mae"
    R2 = "r2"
    MAPE = "mape"
    SMAPE = "smape"
    PRECISION = "precision"
    RECALL = "recall"
    ACCURACY = "accuracy"
    ROCAUC = "roc_auc"
    F1 = "f1"
    LOGLOSS = "log_loss"


class SplitType(str, Enum):
    TRAIN = "train"
    VALI = "validation"
    TEST = "test"


def get_style(continual_style):
    if continual_style is None:
        c = Client()
        continual_style = c.config.style
    if not isinstance(continual_style, str):
        continual_style = continual_style.value
    if continual_style.upper() == "WHITE":
        return continual_style_white
    elif continual_style.upper() == "BLACK":
        return continual_style_black
    elif continual_style.upper() == "BLUE":
        return continual_style_blue
    elif continual_style.upper() == "GREEN":
        return continual_style_green
    elif continual_style.upper() == "PURPLE":
        return continual_style_purple
    elif continual_style.upper() == "DARK_BLUE":
        return continual_style_dark_blue


# wrapper function, mainly for convenience.
# but helps prevent calling get_project and get_environment in an undesirable order
def get_project_and_environment(project, environment):
    return (get_project(project), get_environment(project, environment))


# takes project and org display names
# returns project object
# creates project if it doesn't exist
def get_or_create_project(project_name, org_name, data_store, project_id=None):
    # Associate type string with a data store type
    data_store_map = {
        "snowflake": types.DataStoreSnowflake,
        "big_query": types.DataStoreBigQuery,
        "redshift": types.DataStoreRedshift,
        "postgres": types.DataStorePostgres,
    }
    type_name = data_store["type"]

    del data_store["type"]

    data_store_args_dict = {name: None for name in data_store_map}
    data_store_args_dict[type_name] = data_store_map[type_name](**data_store)

    # get project or create if does not exist
    c = Client()
    try:
        if not project_id:
            project_id = "projects/%s" % project_name
            project = get_project_from_display_name(project_name)
        else:
            project = c.projects.get(project_id)
    except Exception:
        org1 = get_or_create_org(org_name)
        ds = types.DataStore(type=type_name, **data_store_args_dict)
        project = c.projects.create(
            display_name=project_name,
            project_id=project_id,
            data_store=ds,
            organization=org1.name,
        )
    return project


# takes environment display names
# returns environment object
# creates environment if it doesn't exist
def get_or_create_environment(display_name, data_store, environment_name=None):
    c = Client()
    try:
        if not environment_name:
            env = get_environment_from_display_name(display_name)
        else:
            env = c.projects.get(environment_name)
    except Exception:
        env = c.environments.create(
            display_name=display_name,
            data_store=data_store,
        )
    return env


# returns proj obj of given display_name
# returns first found, if multiple exist w/ same name
# this exists because you can't re-use a project id once assigned,
### so the previous function may change your project id from projects/<project_name>
### to projects/<project_name>-#. This can be unexpected and lead to errors when trying
### to access the project downstream
# essentially, this imposes uniqueness of the display names
def get_project_from_display_name(project_name):
    c = Client()
    try:
        projs = c.projects.list()
        proj = [x for x in projs if x.display_name == project_name][0]
        return proj
    except:
        raise


def get_environment_from_display_name(project_name, environment_name):
    c = Client()
    try:
        envs = c.environments.list()
        env = [x for x in envs if x.display_name == environment_name][0]
        return env
    except:
        raise


# takes org display name or optionally id
# returns org object
# creates org if display_name doesn't exist
def get_or_create_org(display_name, id=None):
    c = Client()
    try:
        if not id:
            org = get_org_from_display_name(display_name)
        else:
            org = c.organizations.get(id)
    except Exception:
        org = c.organizations.create(display_name)
    return org


# quick fix for dev-464
# returns org id of given display_name
# returns first found, if multiple exist w/ same name
def get_org_from_display_name(org_name):
    c = Client()
    try:
        orgs = c.organizations.list()
        org = [x for x in orgs if x.display_name == org_name][0]
        return org
    except:
        raise


def get_project(project):
    if (project is None) or (project == ""):
        project = get_default_project()
    else:
        if "@" in project:
            project = project.split("@")[0]
    return project


def get_default_project():
    c = Client()
    proj = c.config.project
    if proj is None or proj == "":
        typer.secho(
            "No project set.  Use continual config set-project or --project to set a project.",
            fg="red",
        )
        exit(1)
    return proj.split("/")[-1]


##get environment
# first priority is to use environment variable if it's passed
# second, try to get env out of project string via project@env (either passed or via default project)
# third, get env from git branch
# 4th, use prod (default)
# note: master/main/production are all mapped to 'prod'
def get_environment(project, environment):
    if environment is None or environment == "":
        if project is not None and "@" in project:
            environment = project.split("@")[1]
        else:
            environment = get_default_environment()
    if (
        (environment.lower() == "master")
        or (environment.lower() == "main")
        or (environment.lower() == "production")
    ):  # prevent use of main/master/production, whether we got from git or passed in manually
        environment = "prod"
    if "@" in environment:
        environment = environment.split("@")[-1]
    return environment


##get default environment
# if no environment passed...
# check to see if config has it set.
# if not, get it from git branch
#  if no git branch, then use prod
def get_default_environment():
    c = Client()
    env = c.config.environment
    if env is None or env == "":
        try:
            repo = Repo(search_parent_directories=True)
            env = repo.active_branch.name
        except:  # if not in a git repo, just use prod as default: Danger! Danger! High Voltage!
            pass
    if env is None or env == "":
        env = "prod"
    return env


def get_environment_name(project, environment) -> str:
    project = get_project(project)
    environment = get_environment(project, environment)
    if not project.startswith("projects/"):
        project = f"projects/{project}"
    return f"{project}@{environment}"


def get_endpoint():
    c = Client()
    endpoint = c.config.endpoint
    # Assume development with SSH port forwarding.
    if endpoint == "http://localhost" or endpoint == "http://host.docker.internal":
        endpoint = "http://localhost:9999"
    else:
        endpoint = "https://cloud.continual.ai"
    return endpoint


def print_table(
    data,
    headers,
    print=True,
    style=continual_style_green,
    row_styles=["default", "dim"],
    title=None,
    caption=None,
):
    table = Table(
        expand=True, style=style, row_styles=row_styles, title=title, caption=caption
    )
    for x in headers:
        table.add_column(x, overflow="fold")
    if isinstance(data, pd.DataFrame):
        for i, row in data.iterrows():
            table.add_row(*[str(x) for x in row])
    else:
        for row in data:
            table.add_row(*[str(x) for x in row])
    if print:
        console = Console()
        console.print(table)
    return table


def print_info(
    data, show_index=False, headers="", tablefmt="plain", style=continual_style_blue
):
    # typer.echo(tabulate(data, headers=headers, showindex=show_index, tablefmt=tablefmt))
    table = Table(style=style).grid(expand=True)
    for x in headers:
        table.add_column(x, overflow="fold")
    if isinstance(data, pd.DataFrame):
        for i, row in data.iterrows():
            table.add_row(*[str(x) for x in row])
    else:
        for row in data:
            table.add_row(*[str(x) for x in row])
    console = Console()
    console.print(table)


def process_template(template, layout, split_direction):
    children = []
    next_split = {}
    for obj in template:
        data = obj.get("data", None)
        split = obj.get("children", None)
        if data is not None:
            children.append(
                Layout(
                    obj.get("data"),
                    name=obj.get("name"),
                    ratio=obj.get("ratio", 1),
                    minimum_size=5,
                )
            )
        elif split is not None:
            name = obj.get("name")
            direction = obj.get("direction", "horizontal")
            children.append(
                Layout(name=name, ratio=obj.get("ratio", 1), minimum_size=5)
            )
            next_split["%s:%s" % (name, direction)] = split

    if len(children) > 0:
        layout.split(direction=split_direction, *children)

    if len(next_split) > 0:
        for index in next_split:
            parts = index.split(":")
            name = parts[0]
            direction = parts[1]
            process_template(next_split[index], layout[name], direction)


# template is a list of dict. Each dict defines a split or renderable
# renderable: [{"name":"Top", "ratio": 1, data:<renderable>},
# split: {"name":"Bottom", "ratio": 1, "direction": "horizontal", children: [<next_template]}]
#
def build_dashboard(template, direction="horizontal"):
    layout = Layout()
    process_template(template, layout, direction)
    return layout


def execute_cmd(cmd, print_it=True):
    popen = subprocess.Popen(
        [cmd], stdout=subprocess.PIPE, universal_newlines=True, shell=True
    )
    output = popen.stdout.read()
    exit_code = popen.wait()
    if print_it:
        print(output)
        # print("exit code: %s" %exit_code)
    return (output, exit_code)


def split_large_csv(csv, output, max_chunk_size=209715200):
    if not os.path.exists(output):
        os.mkdir(output)
    execute_cmd(
        "split -C %s -d %s %s/%s" % (max_chunk_size, csv, output, csv.split("/")[-1])
    )
    return [file.path for file in os.scandir(output) if file.is_file()]


# for progress bar


def chunker(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def exit_with_error(e):
    """Print error and exit.

    Args:
        e: An exception to print.
    """
    if isinstance(e, BaseException):
        typer.secho("Error: " + str(e.message), fg="red")
        if isinstance(e.details, dict) and len(e.details) > 0:
            typer.secho("Details:", fg="red")
            for key, value in e.details.items():
                typer.secho(f"  {key}: {value}", fg="red")
    else:
        typer.secho("Error: " + str(e), fg="red")
    exit(1)


def exit_on_error(func):
    """Decorator that prints and exits on exceptions.

    Args:
        func: Function to wrap.
    Returns:
        A function that prints error and exit(1) on any exception.
    """

    @wraps(func)
    def function_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.debug("an error occured in command", exc_info=e)
            exit_with_error(e)

    return function_wrapper


# formats yaml in more 'user-friendly' ordering, as the default ordering returned by API is alphabetical
def format_yaml(old_yaml):

    type = old_yaml.get("type")
    name = old_yaml.get("name")
    new_schema = {
        "type": type,
        "name": name,
    }

    description = old_yaml.get("description", None)
    if description:
        new_schema["description"] = description

    owners = old_yaml.get("owners", None)
    if owners:
        new_schema["owners"] = owners

    new_schema["index"] = old_yaml.get("index")

    time_index = old_yaml.get("time_index", None)
    if time_index:
        new_schema["time_index"] = time_index

    entity = old_yaml.get("entity", None)
    if entity:
        new_schema["entity"] = entity

    target = old_yaml.get("target", None)
    if target:
        new_schema["target"] = target

    split = old_yaml.get("split", None)
    if split:
        new_schema["split"] = split

    columns = old_yaml.get("columns", None)
    if columns:
        new_schema["columns"] = columns

    exclude_columns = old_yaml.get("exclude_columns", None)
    if exclude_columns:
        new_schema["exclude_columns"] = exclude_columns

    profile = old_yaml.get("profile", None)
    if profile:
        new_schema["profile"] = profile

    train = old_yaml.get("train", None)
    if train:
        new_schema["train"] = train

    predict = old_yaml.get("predict", None)
    if predict:
        new_schema["predict"] = predict

    promote = old_yaml.get("promote", None)
    if promote:
        new_schema["promote"] = promote

    table = old_yaml.get("table", None)
    if table:
        new_schema["table"] = table

    query = old_yaml.get("query", None)
    if query:
        new_schema["query"] = (
            query.replace("\n", " ").replace("\t", " ").replace("\\", " ")
        )  # remove stuff that breaks query

    return yaml.safe_dump(new_schema, sort_keys=False)


# takes list of yamls tuples (name,text) and saves them to file
def save_yamls(yamls, continual_dir, print_it=True):
    paths = []
    if not os.path.exists(continual_dir):
        os.makedirs(continual_dir)
    for yaml in yamls:
        path = "%s/%s.yaml" % (continual_dir, yaml[0])
        with open(path, "w+") as f:
            if print_it:
                typer.secho("Saving Continual yaml definition %s!" % path)
            f.write(yaml[1])
        paths.append(path)
    return paths


def print_change(push, project_id, environment_id, message, include_estimate=False):
    project_env_id = ProjectEnvironmentIdentifer(project_id, environment_id)
    typer.secho(f"Updating {project_env_id.environment_name_short}:", fg="magenta")
    typer.echo(f"  {message or ''}\n")
    name = push.name
    typer.secho(
        "Operations for %s in project '%s', environment '%s':"
        % (push.id, project_env_id.project_id, project_env_id.environment_id_alias),
        fg="magenta",
    )
    if len(push.plan) > 0:
        steps = sorted(
            push.plan,
            key=lambda x: int(x.id),
            reverse=False,
        )
        typer.secho(
            f"\n  {'Operation':20s}{'State':20s}",
            fg="magenta",
        )
        old_resource = ""
        featuresets = []
        models = []
        # the id ordering we get back from continual is kind of wonky and has models listed first
        credits = {}
        for step in steps:
            if step.resource_name.split("/")[2] == "models":
                models.append(step)
            else:
                featuresets.append(step)
        for step in featuresets + models:
            step_id = step.resource_name.split("/")[-1]
            if step.resource_name.split("/")[2] == "models":
                step_type = "Model"
            else:
                step_type = "Feature Set"
            if not (old_resource == step_id):
                typer.secho("\n  %s: %s" % (step_type, step_id), fg="blue")
                old_resource = step_id
            typer.echo(f"  {step.operation:20s}{step.state:20s}")
            if include_estimate:
                if step.operation.value == "TRAIN":
                    credits_used = get_model_credit_estimate(step.resource_name)
                    credits["%s %s" % (step_id, step.operation.value)] = credits_used
                if step.operation.value == "PREDICT":
                    credits_used = get_batch_prediction_credit_estimate(
                        step.resource_name
                    )
                    credits["%s %s" % (step_id, step.operation.value)] = credits_used

        # process schedules
        push_schedule = {}
        for step in models:
            if step.operation.value in ["CREATE", "UPDATE"]:
                schema_text = yaml.load(step.schema_text, Loader=yaml.FullLoader)
                train_text = schema_text.get("train")
                if train_text:
                    schedule = train_text.get("schedule")
                    if schedule:
                        push_schedule[
                            "%s %s" % (step.resource_name.split("/")[-1], "TRAIN")
                        ] = schedule
                predict_text = schema_text.get("predict")
                if predict_text:
                    schedule = predict_text.get("schedule")
                    if schedule:
                        push_schedule[
                            "%s %s" % (step.resource_name.split("/")[-1], "PREDICT")
                        ] = schedule

        if len(push_schedule) > 0:
            print_push_scheules(push_schedule)

        if include_estimate and len(credits) > 0:
            print_credit_estimate_summary(credits)

    else:
        typer.secho("\n  No changes detected.", fg="red")

    typer.secho("\nLink:", fg="magenta")
    push_id = name.split("/")[-1]
    endpoint = get_endpoint()

    typer.secho(
        f"  {endpoint}/{project_env_id.url_path}/changes/{push_id}\n",
        fg="blue",
    )


def get_model_credit_estimate(model, project=None, environment=None, n=100):
    project, environment = get_project_and_environment(project, environment)
    c = Client(project=get_environment_name(project, environment))
    credits = "N/A"

    # get last successful mv. This should usually be latest_model_version, but there are cases where this won't happen.
    try:
        m = c.models.get(model)
        mv = c.model_versions.get(m.latest_model_version)
        if mv.state.value != "SUCCEEDED":
            # get last successful model version, resources are returned in reverse order, so take last one
            mv = m.model_versions.list(1000, filters=["state:SUCCEEDED"])[-1]
        credits = mv.credits_used
        if credits == 0:
            credits = "N/A"
    except:  # if we erorr out just return "N/A"
        pass

    return credits


def get_batch_prediction_credit_estimate(model, project=None, environment=None):
    project, environment = get_project_and_environment(project, environment)
    c = Client(project=get_environment_name(project, environment))
    credits = "N/A"

    # get last successful bpj. This should usually be latest_batch_prediction, but there are cases where this won't happen.
    try:
        m = c.models.get(model)
        bpj = c.batch_prediction_jobs.get(m.latest_batch_prediction)
        if bpj.state.value != "SUCCEEDED":
            # get last successful prediction, resources are returned in reverse order, so take last one
            bpj = m.batch_prediction_jobs.list(1000, filters=["state:SUCCEEDED"])[-1]
        credits = bpj.credits_used
        if credits == 0:
            credits = "N/A"
    except:  # if we hit an error, just return "N/A"
        pass

    return credits


def print_push_scheules(schedule):
    typer.secho(
        "\nSchedules for changed models: ",
        fg="magenta",
    )
    typer.secho(
        f"\n  {'Model':40s}{'Operation':20s}{'Schedule':20s}",
        fg="magenta",
    )
    for resource, cron in schedule.items():
        [model, op] = resource.split(" ")
        try:
            cron_pretty = get_description(cron)
        except:
            cron_pretty = cron
        typer.secho(f"  {model:40s}{op:20s}{cron_pretty}", fg="white")


def print_credit_estimate_summary(credits):
    estimates = {k: v for (k, v) in credits.items() if v != "N/A"}
    total_credit_usage = sum(estimates.values())
    if total_credit_usage == 0:
        total_credit_usage = "N/A"
    typer.secho(
        "\nTotal estimated credits to be used: %s " % total_credit_usage,
        fg="magenta",
    )
    typer.secho(
        f"\n  {'Resource':40s}{'Operation':20s}{'Estimated Credits':20s}",
        fg="magenta",
    )
    for operation, estimate in credits.items():
        [res, op] = operation.split(" ")
        typer.secho(f"  {res:40s}{op:20s}{estimate}", fg="white")

    if total_credit_usage == "N/A":
        typer.secho(
            f"\n  Unable to find any previous credit usage for affected reources. \n  You must train a model or run a batch prediction before estimates can be made.",
            fg="magenta",
        )


# print a push plan
def print_change_plan(push, project, environment):

    typer.secho(
        "Operations for %s in project %s, environment %s:"
        % (push.id, project, environment),
        fg="magenta",
    )
    typer.secho("\t" + push.message)
    if len(push.plan) > 0:
        steps = sorted(
            push.plan,
            key=lambda x: int(x.id),
            reverse=False,
        )
        typer.secho(
            f"\n  {'Operation':20s}{'State':20s}{'Duration(s)':12s}{'Start Time':30s}{'End Time':30s}",
            fg="magenta",
        )
        old_resource = ""
        featuresets = []
        models = []
        for step in steps:
            if step.resource_name.split("/")[2] == "models":
                models.append(step)
            else:
                featuresets.append(step)
        for step in featuresets + models:
            step_id = step.resource_name.split("/")[-1]
            if step.resource_name.split("/")[2] == "models":
                step_type = "Model"
            else:
                step_type = "Feature Set"
            start_time = step.start_time
            end_time = step.finish_time
            if start_time:
                start_time = step.start_time.replace(microsecond=0)
                if end_time:
                    end_time = end_time.replace(microsecond=0)
                    duration = (end_time - start_time).seconds
                else:
                    end_time = "N/A"
                    duration = "N/A"
            else:
                start_time = "N/A"
                end_time = "N/A"
                duration = "N/A"
            if not (old_resource == step_id):
                typer.secho("\n  %s: %s" % (step_type, step_id), fg="blue")
                old_resource = step_id
            typer.echo(
                f"  {step.operation:20s}{step.state:20s}{str(duration):12s}{str(start_time):30s}{str(end_time):30s}"
            )
    else:
        typer.secho("\n  No changes found.", fg="red")


def get_message(message, default):
    if not message:
        message = default
    if not message:
        try:
            repo = Repo(search_parent_directories=True)
            message = repo.git.log(n=1, oneline=True)
            if repo.is_dirty():
                message = message.split(" ")
                message[0] = message[0] + "+dirty"
                message = " ".join(message)
        except Exception:
            pass
    return message


def get_datastore_database_and_schema(data_store):
    if data_store.type == "":
        return ("", "")

    fs_type = data_store.type

    if fs_type == "bigquery":
        db_schema = data_store.__getattribute__("big_query").dataset
        database = ""
    elif fs_type == "databricks":
        db_schema = ""
        database = data_store.__getattribute__(fs_type).database
    else:
        db_schema = data_store.__getattribute__(fs_type).db_schema
        database = data_store.__getattribute__(fs_type).database

    return (db_schema, database)



# Public SDK interface.
from continual.python.sdk.client import Client
from continual.python.sdk import exceptions
from continual.python.sdk.users import User
from continual.python.sdk.organizations import Organization
from continual.python.sdk.projects import Project
from continual.python.sdk.featuresets import FeatureSet
from continual.python.sdk.models import Model
from continual.python.sdk.model_versions import ModelVersion
from continual.python.sdk.experiments import Experiment
from continual.python.sdk.ingestions import Ingestion
from continual.python.sdk.promotions import Promotion

# TODO: Figure out what we want to do with nested types.
from continual.rpc.management.v1.types import (
    IngestionState,
    ModelVersionState,
    PromotionState,
)

continual = Client()
"""Client singleton.

To instantiate a custom client use `continual.Client()`.
"""

__all__ = [
    "continual",
    "Client",
    "exceptions",
    "User",
    "Organization",
    "Project",
    "FeatureSet",
    "Connection",
    "Model",
    "ModelVersion",
    "Experiment",
    "Ingestion",
    "Promotion",
    "IngestionState",
    "ModelVersionState",
    "PromotionState",
]



from continual.python.sdk.exceptions import InvalidArgumentError
from continual.python.utils.utils import parse_feature_set_name
import copy
from google.protobuf.json_format import MessageToDict
from continual.rpc.management.v1 import management_types_pb2

MAX_ENTITY_DEPTH = 5


def cleanup_value(qd):
    for f in qd["fields"]:
        if "value" in f:
            del f["value"]
        if "relation" in f:
            cleanup_value(f["relation"])

    if "related" in qd:
        for q in qd["related"]:
            cleanup_value(q)


class Query:
    def __init__(self, qrepr):
        if isinstance(qrepr, management_types_pb2.Query):
            self._qrepr = qrepr
        elif isinstance(qrepr, bytes):
            self._qrepr = management_types_pb2.Query()
            self._qrepr.ParseFromString(qrepr)
        elif isinstance(qrepr, str):
            qrepr_b = eval(qrepr)
            self._qrepr = management_types_pb2.Query()
            self._qrepr.ParseFromString(qrepr_b)
        else:
            self._qrepr = management_types_pb2.Query(**qrepr)

        self._inputs = None
        self._outputs = None
        self._featuresets = None

    def serialize(self):
        return self._qrepr.SerializeToString()

    def to_proto(self):
        return self._qrepr

    def to_dict(self):
        return MessageToDict(self._qrepr, preserving_proto_field_name=True)

    def __repr__(self):
        return repr(self._qrepr)

    def set_id(self, id):
        for f in self._qrepr.fields:
            if f.index == True:
                if isinstance(id, str):
                    f.value.text = id
                else:
                    f.value.number = id
                return

        id_feature = management_types_pb2.QueryFeature(
            name="id",
            type=management_types_pb2.FieldType.TEXT,
            model=False,
            value=management_types_pb2.FeatureValue(text=id),
        )

        self._qrepr.fields.append(id_feature)

    @property
    def fields(self):
        return self._qrepr.fields

    @property
    def related(self):
        return self._qrepr.related

    def clone(self):
        return Query(copy.deepcopy(self._qrepr))

    def build_field_list(
        self, query, path, inputs, outputs, featuresets, include_id, related=False
    ):
        id = query.name.split("/")[3]

        if path is None:
            featuresets.append(
                {"name": id, "type": management_types_pb2.FieldType.TEXT}
            )

        for f in query.fields:
            if f.exclude:
                continue

            # Skip index for related featuresets.
            if related and f.type == management_types_pb2.FieldType.INDEX:
                continue

            # For models the field names are not qualified.
            if query.name.split("/")[2] == "models":
                qualified_name = f.name
            else:
                qualified_name = id + ":" + f.name
            if path is not None:
                qualified_name = path + "$" + qualified_name
            if f.HasField("relation"):
                featuresets.append({"name": qualified_name, "type": f.type})
                self.build_field_list(
                    f.relation, qualified_name, inputs, outputs, featuresets, include_id
                )
            else:
                if f.model:
                    outputs.append({"name": qualified_name, "type": f.type})
                else:
                    if f.index and not include_id:
                        continue

                    if f.time_index and not include_id:
                        continue

                    if path is not None and f.index:
                        inputs.append(
                            {
                                "name": path,
                                "type": f.type,
                                "index": f.index,
                                "time_index": f.time_index,
                            }
                        )
                    else:
                        inputs.append(
                            {
                                "name": qualified_name,
                                "type": f.type,
                                "index": f.index,
                                "time_index": f.time_index,
                            }
                        )

        for q in query.related:
            self.build_field_list(
                q, path, inputs, outputs, featuresets, include_id, related=True
            )

    def get_model_version_io(self, include_id=True):
        if (
            self._inputs is not None
            and self._outputs is not None
            and self._featuresets is not None
        ):
            return self._inputs, self._outputs
        inputs = []
        outputs = []
        featuresets = []
        query = self._qrepr
        self.build_field_list(query, None, inputs, outputs, featuresets, include_id)
        self._inputs = inputs
        self._outputs = outputs
        self._featuresets = featuresets
        return inputs, outputs

    def can_hydrate(self, overrides: dict):

        has_hydrateable_values = False

        inputs, outputs = self.get_model_version_io(include_id=True)

        has_missing_values = False
        for input in inputs:
            input_name = input["name"]
            if input_name in overrides:
                overrides[input_name]
                if input["index"]:
                    has_hydrateable_values = True
            else:
                if not input["index"]:
                    has_missing_values = True

        return has_hydrateable_values and has_missing_values

    def _override(self, query, key_splits, value):
        splits = key_splits[0].split(":")

        fieldId = key_splits[0]
        featureset_id = None
        if len(splits) > 1:
            featureset_id = splits[0]
            fieldId = splits[1]

        comps = parse_feature_set_name(query["name"])
        resource_id = ""
        if "featureSets" in comps:
            resource_id = comps["featureSets"]
        elif "models" in comps:
            resource_id = comps["models"]
        if featureset_id is not None and resource_id != featureset_id:
            raise InvalidArgumentError(
                "error - path of override [{}] does not match query for name [{}]".format(
                    key_splits[0], resource_id
                )
            )

        for i in range(0, len(splits)):
            for f in query["fields"]:
                if f["name"] != fieldId:
                    continue

                if len(key_splits) == 1:
                    f["value"] = value
                    return

                if f["relation"] != None:
                    self._override(f["relation"], key_splits[1:], value)

        return

    def override(self, overrides: dict):
        qdict = self.to_dict()
        cleanup_value(qdict)
        for key in overrides:
            value = overrides[key]
            splits = key.split("$")
            current_fs = qdict
            self._override(qdict, splits, value)
            for i in range(0, len(splits)):
                for f in current_fs["fields"]:
                    if f["name"] != splits[i]:
                        continue

                    if len(splits) == 1:
                        f["value"] = value
                        continue

                    if f["relation"] is not None:
                        current_fs = f["relation"]
        return qdict

    def get_index_name(self):
        for f in self._qrepr.fields:
            if f.type == management_types_pb2.FieldType.INDEX or f.index:
                return f.name
        return None

    def get_split_name(self):
        for f in self._qrepr.fields:
            if f.type == management_types_pb2.FieldType.SPLIT or f.split:
                return f.name
        return None

    def get_time_index_name(self):
        for f in self._qrepr.fields:
            if f.type == management_types_pb2.FieldType.TIME_INDEX or f.time_index:
                return f.name
        return None

    def get_exclude_list(self):
        exclude_fields = []
        for f in self._qrepr.fields:
            if f.exclude:
                exclude_fields.append(f.qualified_name)

            if f.relation is not None:
                r_excludes = Query(f.relation).get_exclude_list()
                exclude_fields.extend(r_excludes)

        for r in self._qrepr.related:
            r_excludes = Query(r).get_exclude_list()
            exclude_fields.extend(r_excludes)
        return exclude_fields

    def get_dtype(self, field_name: str):
        fs_splits = field_name.split("$")
        """
             "id"  ==> Model field.
             "tweets:text"  ==> Related entity field.
             "author$users:name" ==> Joined relation field.
             "author:$user:location$locations:name" ==> Nested join field.
        """
        object_name = None
        if ":" in fs_splits[0]:
            column_name = fs_splits[0].split(":")[1]
            object_name = fs_splits[0].split(":")[0]
        else:
            column_name = fs_splits[0]

        # Check if the name can be found in fields.
        for f in self._qrepr.fields:
            if column_name == f.name:
                if len(fs_splits) > 1 and f.relation is not None:
                    return Query(f.relation).get_dtype("$".join(fs_splits[1:]))
                else:
                    return f.dtype

        if object_name is None:
            return ""
        for r in self._qrepr.related:
            if r.name.split("/")[3] == object_name:
                if len(fs_splits) > 1:
                    return Query(r).get_dtype("$".join(fs_splits[1:]))
                else:
                    return Query(r).get_dtype(field_name)
        return ""

    def get_field(self, field_name: str):
        fs_splits = field_name.split("$")
        """
             "id"  ==> Model field.
             "tweets:text"  ==> Related entity field.
             "author$users:name" ==> Joined relation field.
             "author:$user:location$locations:name" ==> Nested join field.
        """
        object_name = None
        if ":" in fs_splits[0]:
            column_name = fs_splits[0].split(":")[1]
            object_name = fs_splits[0].split(":")[0]
        else:
            column_name = fs_splits[0]

        # Check if the name can be found in fields.
        for f in self._qrepr.fields:
            if column_name == f.name:
                if len(fs_splits) > 1 and f.relation is not None:
                    return Query(f.relation).get_dtype("$".join(fs_splits[1:]))
                else:
                    return f

        if object_name is None:
            return None

        for r in self._qrepr.related:
            if r.name.split("/")[3] == object_name:
                if len(fs_splits) > 1:
                    return Query(r).get_dtype("$".join(fs_splits[1:]))
                else:
                    return Query(r).get_dtype(field_name)
        return None

    @staticmethod
    def _contains_parent_id(fs_ids: dict, name):
        parent = name
        while ":" in parent:
            parent = parent[0 : parent.rindex(":")]
            if "$" in parent:
                parent = parent[0 : parent.rindex("$")]
            if parent in fs_ids:
                return True

        return False

    @staticmethod
    def deserialize(encoded):
        nqp = management_types_pb2.Query()
        nqp.ParseFromString(encoded)
        return Query(nqp)

    @staticmethod
    def get_related_feature_sets(entity: str, featuresets: dict):
        ret_list = []
        for fs in featuresets:
            featureset = featuresets[fs]
            if featureset.schema.entity != entity:
                continue
            ret_list.append(featureset)
        return ret_list

    @staticmethod
    def form_query(
        query: dict = None,
        model=None,
        normalized: bool = False,
        related: bool = False,
        metadata: bool = False,
        path=None,
        all_featuresets: dict = None,
        level=0,
    ) -> management_types_pb2.Query:

        """Recursively forms a training data set query."""
        j_dict = query
        if j_dict is None:
            j_dict = management_types_pb2.Query()

        # Start with level 1.
        # MAX_ENTITY_DEPTH defines the maximum recursion of joins to do.
        level = level + 1

        is_model = False

        if model.name.split("/")[2] == "models":
            is_model = True

        model_id = model.name.split("/")[3]
        j_dict.name = model.name

        if not is_model:
            j_dict.entity_id = model.schema.entity
        if model.schema.table:
            j_dict.table_name = model.schema.table

        exclude_columns = []
        if model.schema.exclude_columns is not None:
            exclude_columns = set(model.schema.exclude_columns)

        for m in model.schema.columns:
            # TODO remove the metadata features if not requested.
            # if not metadata and not m.index and not m.time_index:
            #    continue
            qualified_name = m.name
            if related:
                qualified_name = model_id + ":" + m.name

            if path is not None:
                qualified_name = model_id + ":" + m.name
                qualified_name = path + "$" + qualified_name

            # TODO: track metadata field.
            qfeature = management_types_pb2.QueryFeature(
                name=m.name,
                type=m.type,
                dtype=m.dtype,
                qualified_name=qualified_name,
            )
            if is_model and m.name == model.schema.target:
                qfeature.model = True

            if (
                qfeature.type == management_types_pb2.INDEX
                or m.name == model.schema.index
            ):
                qfeature.index = True

            if (
                qfeature.type == management_types_pb2.TIME_INDEX
                or m.name == model.schema.time_index
            ):
                qfeature.time_index = True

            if qfeature.type == management_types_pb2.SPLIT or (
                is_model and m.name == model.schema.split
            ):
                qfeature.split = True

            if m.name in exclude_columns:
                qfeature.exclude = True

            if m.entity is not None and m.entity != "" and level < MAX_ENTITY_DEPTH:
                fss = Query.get_related_feature_sets(m.entity, all_featuresets)
                if fss == None or len(fss) == 0:
                    continue

                Query.form_query(
                    qfeature.relation,
                    fss[0],
                    None,
                    metadata=metadata,
                    path=qualified_name,
                    all_featuresets=all_featuresets,
                    level=level,
                )

            j_dict.fields.append(qfeature)
        if not related and not is_model and model.schema.entity is not None:
            all_fs = Query.get_related_feature_sets(
                model.schema.entity, all_featuresets
            )
            for fs in all_fs:
                if fs.name == model.name:
                    continue
                sub_query = management_types_pb2.Query()
                Query.form_query(
                    sub_query,
                    fs,
                    None,
                    related=True,
                    metadata=metadata,
                    path=path,
                    all_featuresets=all_featuresets,
                    level=level,
                )
                j_dict.related.append(sub_query)

        return Query(j_dict)



import os
from pathlib import Path
import yaml
from typing import Optional


class Config:
    """Config for client.

    Reads configuration from `~/.continual/continual.yaml` or `CONTINUAL_CONFIG` if
    not set during initialization.  Priority of config variables is `__init__`,
    environmental variables, and then the configuration file.

    Environmental variables are:

    * `CONTINUAL_CONFIG` - Configuration file.
    * `CONTINUAL_API_KEY` - API Key.
    * `CONTINUAL_PROJECT` - Default project.
    * `CONTINUAL_ENVIRONMENT` - Default environment.
    * `CONTINUAL_ENDPOINT` - Cluster endpoint.
    * `CONTINUAL_STYLE` - Continual style
    """

    email: str = ""
    """Email."""

    api_key: str = ""
    """API key."""

    project: str = ""
    """Default project."""

    environment: str = ""
    """Default environment."""

    endpoint: str = "https://api.continual.ai"
    """Cluster endpoint."""

    _configfile: str

    style: str = ""
    """Default template style"""

    def __init__(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        project: Optional[str] = None,
        environment: Optional[str] = None,
        endpoint: Optional[str] = None,
        style: Optional[str] = None,
    ) -> None:
        """Initialize config.

        Args:
            api_key: API key.
            email: Email address.
            project: Default project.
            endpoint: Cluster endpoint.
            style: Continual style.
        """
        self._configfile = os.path.join(os.getcwd(), ".continual.yaml")
        if not os.path.exists(self._configfile):
            self._configfile = os.environ.get(
                "CONTINUAL_CONFIG",
                os.path.join(Path.home(), ".continual", "continual.yaml"),
            )
        config = dict()
        if os.path.exists(self._configfile):
            with open(self._configfile, "r") as stream:
                config = yaml.safe_load(stream) or dict()
        self.api_key = (
            api_key or os.environ.get("CONTINUAL_API_KEY") or config.get("api_key")
        )
        self.project = (
            project or os.environ.get("CONTINUAL_PROJECT") or config.get("project")
        )
        if self.project is not None and self.project != "" and "/" not in self.project:
            self.project = "projects/" + self.project
        env_from_proj = None
        if self.project is not None and "@" in self.project:
            # don't overwrite environment if it was explicity set in its own field.
            # Ideally this shouldn't ever really happen
            env_from_proj = self.project.split("@")[1]
            self.project = self.project.split("@")[0]
        self.environment = (
            environment
            or env_from_proj
            or os.environ.get("CONTINUAL_ENVIRONMENT")
            or config.get("environment")
        )

        self.endpoint = (
            endpoint
            or os.environ.get("CONTINUAL_ENDPOINT")
            or config.get("endpoint", "https://api.continual.ai")
        )
        self.email = email or config.get("email")

        self.style = style or config.get("style") or os.environ.get("CONTINUAL_STYLE")
        if self.style is None:
            self.style = "GREEN"

    def set_project(self, project: str, save=True) -> str:
        if project and "/" not in project:
            project = "projects/" + project

        if "@" in project:
            self.environment = project.split("@")[1]
            self.project = project.split("@")[0]
        else:
            self.project = project

        if save:
            self.save()

        return self.project

    def set_environment(self, environment: str, save=True) -> str:
        if environment and "@" in environment:
            environment = environment.split("@")[-1]

        self.environment = environment
        if save:
            self.save()

        return self.environment

    def show(self):
        """Print config information."""
        print(f"Config File:", self._configfile)
        print(f"Email: {self.email or ''}")
        print(f"Endpoint: {self.endpoint or ''}")
        print(f"Project: {self.project or ''}")
        print(f"Environment: {self.environment or ''}")
        print(f"Style: {self.style or ''}")
        if self.api_key is None or self.api_key == "":
            print("API Key:")
        else:
            print("API Key: ****************")

    def save(self):
        """Save config to config file."""
        Path(os.path.dirname(self._configfile)).mkdir(parents=True, exist_ok=True)
        with open(self._configfile, "w") as outfile:
            config = dict(
                email=self.email,
                api_key=self.api_key,
                project=self.project,
                environment=self.environment,
                endpoint=self.endpoint,
                style=self.style,
            )
            yaml.dump(config, outfile)



from halo import Halo
import time
import math
from typing import List, Callable, TypeVar, Optional
from continual.python.sdk.exceptions import FailedError

RESTING_STATES = set(["ONLINE", "OFFLINE", "CANCELLED", "FAILED", "SUCCEEDED"])
FAILED_STATES = set(["FAILED"])

"""Common resting states."""

T = TypeVar("T")


def wait(
    getter: Callable[[], T],
    states: Optional[List[str]] = None,
    echo: bool = False,
    timeout: Optional[int] = None,
) -> T:
    """Wait for state transition.

    Arguments:
        getter: Get resource callable.
        states: Resting state.
        echo: Show progress bar.
        timeout: Maximum time to wait.
    Returns:
        A resource.
    """

    if states is None:
        states = RESTING_STATES
    if timeout is None:
        timeout = math.inf

    spinner = Halo(text="Loading", spinner="dots")
    total_time = 0
    spinner.start()

    while True:
        time.sleep(5)
        resource = getter()
        state = str(resource.state)
        spinner.text = state
        total_time += 5
        if total_time > timeout:
            raise TimeoutError(f"Timeout occured after {total_time} seconds.")
        if state in states:
            break

    if state in ["OFFLINE", "ONLINE", "SUCCEEED"]:
        spinner.succeed()
    elif state in ["CANCELLED", "FAILED"]:
        spinner.fail()
    elif total_time > timeout:
        spinner.text = "TIMEOUT"
        spinner.fail()
    else:
        spinner.stop()
    if state in FAILED_STATES:
        raise FailedError(
            f"{resource.__class__.__name__} transitioned to FAILED state.",
            details=dict(name=resource.name),
        )
    return resource



import typer
import json

from continual.python.cli import utils
from continual import Client

from rich.console import Console
from typing import List

app = typer.Typer(help="Manage experiments.")


def format_experiments_data(e, zipped=False, all_projects=False):
    framework = json.loads(e.training_config).get("framework")
    name_parts = e.name.split("/")
    model = name_parts[3]
    model_version = name_parts[5]
    end_time = e.finish_time
    start_time = e.start_time
    if start_time:
        start_time = start_time.replace(microsecond=0)
    if end_time:
        end_time = end_time.replace(microsecond=0)
        duration = (end_time - start_time).seconds
    else:
        end_time = "N/A"
        duration = "N/A"
    if zipped:
        data = [
            e.id,
            e.name,
            e.state.value,
            framework,
            e.type,
            start_time,
            end_time,
            duration,
            e.performance_metric,
            e.performance_metric_val,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Framework",
            "Algorithm",
            "Start Time",
            "End Time",
            "Duration (s)",
            "Metric",
            "Metric Value",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            e.id,
            e.state.value,
            model,
            model_version,
            # framework,
            e.type.split("/")[-1],
            # start_time,
            # end_time,
            duration,
            e.performance_metric,
            e.performance_metric_val,
        ]
        headers = [
            "ID",
            "State",
            "Model",
            "Model Version",
            # "Framework",
            "Algorithm",
            # "Start Time",
            # "End Time",
            "Duration (s)",
            "Metric",
            "Metric Value",
        ]
        if all_projects:
            data.insert(0, e.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            model_version=None,
            model=None,
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    model_version: str = typer.Option(None, help="Model Version ID."),
    model: str = typer.Option(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(
        30, "--num", "-n", help="Number of records to show.", show_default=False
    ),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List experiments.

    filters can include:
        - state  (i.e. state:FAILED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    model_snippet = ""
    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if model_version is not None:
        c = c.model_versions.get(model_version)
        model_snippet = "for model version %s of model %s " % (
            model_version,
            c.parent.split("/")[-1],
        )
    elif model is not None:
        m = c.models.get(model)
        c = m.model_versions.get(m.latest_model_version)
        model_snippet = "for latest model experiment in model %s " % (model)
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    if all_projects:
        project_snippet = "all accessible projects"
    data = []
    headers = []
    for e in c.experiments.list(n, filters=filters, all_projects=all_projects):
        (e_data, headers) = format_experiments_data(e, all_projects=all_projects)
        data.append(e_data)
    typer.secho(
        "\nFound %s experiments %sin %s%s: "
        % (len(data), model_snippet, project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    experiment: str = typer.Argument(..., help="Experiment ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    json: bool = typer.Option(False, "--json", help="Show full JSON representation."),
):
    """Get experiment details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    e = c.experiments.get(experiment)
    if json:
        console = Console()
        console.print(e.to_dict())
    else:
        data = format_experiments_data(e, zipped=True)
        typer.secho("\nRetrieving experiment %s: \n" % (experiment), fg="blue")
        utils.print_info(data, headers="")



import typer
import os

from continual.rpc.management.v1.types import BatchPredictionState
from continual.python.cli import utils
from continual import Client

from rich.console import Console
from enum import Enum
from pathlib import Path
from typing import List

app = typer.Typer(help="Manage batch predictions.")


class PredictionDestType(str, Enum):
    csv = "CSV"
    store = "STORE"


def format_batch_prediction_data(bpj, zipped=False, all_projects=False):
    name_parts = bpj.name.split("/")
    model = name_parts[3]
    model_version = name_parts[5]
    end_time = bpj.end_time
    start_time = bpj.start_time
    if start_time:
        start_time = start_time.replace(microsecond=0)
    if end_time and start_time:
        end_time = end_time.replace(microsecond=0)
        duration = (end_time - start_time).seconds
    else:
        end_time = "N/A"
        duration = "N/A"
    if zipped:
        data = [
            bpj.id,
            bpj.name,
            bpj.state.value,
            bpj.incremental,
            start_time,
            end_time,
            duration,
            bpj.prediction_count,
            model,
            model_version,
            bpj.prediction_model,
            bpj.table,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Incremental",
            "Start Time",
            "End Time",
            "Duration (s)",
            "Predictions",
            "Model",
            "Model Version",
            "Prediction Model",
            "Data Warehosue Table",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            bpj.id,
            bpj.state.value,
            bpj.incremental,
            model,
            model_version,
            start_time,
            end_time,
            duration,
            bpj.prediction_count,
        ]
        headers = [
            "ID",
            "State",
            "Incremental",
            "Model",
            "Model Version",
            "Start Time",
            "End Time",
            "Duration (s)",
            "Predictions",
        ]
        if all_projects:
            data.insert(0, bpj.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            model=None,
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    model: str = typer.Option(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List batch prediction jobs in a project.

    Filters can include:
        - state  (i.e. state:FAILED)
        - incremental (i.e. incremental: True)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    data = []
    headers = []
    model_snippet = ""
    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if model is not None:
        try:
            c = c.models.get(model)
            model_snippet = "for model %s " % model
        except Exception as e:
            typer.secho(
                "Failed to find model %s in project %s, environment %s. Please double check that your model id, project id, and environment id are correct. : %s"
                % (model, project, environment, str(e)),
                fg="red",
            )
            raise typer.Exit(code=1)
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    if all_projects:
        project_snippet = "all accessible projects & environments"
    for bpj in c.batch_prediction_jobs.list(
        n, filters=filters, all_projects=all_projects
    ):
        (bpj_data, headers) = format_batch_prediction_data(
            bpj, all_projects=all_projects
        )
        data.append(bpj_data)
    typer.secho(
        "\nFound %s batch prediction jobs %sin %s%s:"
        % (len(data), model_snippet, project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    batch_prediction: str = typer.Argument(..., help="Batch prediction ID"),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    project: str = typer.Option(None, help="Project ID"),
    json: bool = typer.Option(False, "--json", help="Show full JSON representation."),
):
    """Get batch prediction job details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    try:
        bpj = c.batch_prediction_jobs.get(batch_prediction)
    except Exception as e:
        typer.secho(
            "Failed to find batch prediction job %s in project %s, environment %s. Please double check that your job id, project id, and environment id are correct. : %s"
            % (batch_prediction, project, environment, str(e)),
            fg="red",
        )
        raise typer.Exit(code=1)
    if json:
        console = Console()
        console.print(bpj.to_dict())
    else:
        data = format_batch_prediction_data(bpj, zipped=True)
        typer.secho(
            "\nRetrieving batch prediction job %s:\n" % batch_prediction, fg="blue"
        )
        utils.print_info(data)


@app.command("run")
@utils.exit_on_error
def run(
    model: str = typer.Argument(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    wait: bool = typer.Option(
        False, "--wait", help="Wait for batch prediction to finish."
    ),
    dest_type: PredictionDestType = typer.Option(
        PredictionDestType.store, help="Destination type.", hidden=True
    ),
    dest_path: Path = typer.Option(
        None, help="Destination path (CSV only)", hidden=True
    ),
    source_type: PredictionDestType = typer.Option(
        PredictionDestType.store, help="Source type.", hidden=True
    ),
    source_path: Path = typer.Option(None, help="Source path (CSV only).", hidden=True),
    debug: bool = typer.Option(False, help="Print full name of batch prediction job."),
    all: bool = typer.Option(False, "--all", help="Run all batch prediction jobs?"),
    state: BatchPredictionState = typer.Option(
        None, help="Only re-run batch predictions job with a certain state."
    ),
    prediction_model: str = typer.Option(None, help="Optional model type override"),
    incremental: bool = typer.Option(
        False, "--incremental", help="Run an incremental prediction job."
    ),
    include_estimate: bool = typer.Option(
        False, "--include-estimate", help="Include credit usage estimate for resources."
    ),
):
    """Runs a batch prediction job for a model.

    If --all or --state is used, it will bulk run all batch prediction jobs
    or all with the given state (i.e --state FAILED will rerun all failed batch prediction jobs)

    --all overrides --state if both are provided.
    """
    if all or state is not None:
        bulk_run(
            project=project,
            environment=environment,
            dest_type=dest_type,
            dest_path=dest_path,
            all=all,
            state=state,
            include_estimate=include_estimate,
        )
    elif model is not None:
        project, environment = utils.get_project_and_environment(project, environment)
        c = Client(project=utils.get_environment_name(project, environment))
        try:
            m = c.models.get(model)
        except Exception as e:
            typer.secho(
                "Failed to find model %s in project %s, environment %s. Please double check that your job id, project id, and environment id are correct. : %s"
                % (model, project, environment, str(e)),
                fg="red",
            )
            raise typer.Exit(code=1)
        if include_estimate:
            credits_used = utils.get_batch_prediction_credit_estimate(
                m.name, project, environment
            )
            typer.secho(
                "Estimated Credit Usage for Batch Prediction: %s\n" % credits_used,
                fg="white",
            )

        bpj = m.batch_predict(
            source_type.value,
            source_path,
            dest_type.value,
            prediction_model=prediction_model,
            incremental=incremental,
        )
        if wait:
            bpj.wait()
            typer.secho(
                "Successfully ran batch prediction job %s for model %s in project %s, environment %s"
                % (bpj.id, m.id, project, environment),
                fg="green",
            )
            if dest_type == PredictionDestType.csv:
                try:
                    os.remove(dest_path)
                except:
                    pass
                bpj.download(dest_path)
                typer.secho(
                    "Downloaded predictions to file %s" % (dest_path), fg="green"
                )
        else:
            typer.secho(
                "Started running batch prediction job %s for model %s in project %s, environment %s\n"
                % (bpj.id, m.id, project, environment),
                fg="green",
            )
            endpoint = utils.get_endpoint()
            p_slug = utils.get_environment_name(project, environment).split("/")[-1]
            typer.secho("You can access the batch prediction job at: ")
            typer.secho(
                f"  {endpoint}/projects/{p_slug}/models/{m.id}/batchPredictions/{bpj.id}\n",
                fg="blue",
            )

        if dest_type == PredictionDestType.store:
            typer.secho("Predictions can be found in feature store at")
            typer.secho("   %s" % (bpj.table), fg="blue")
        if debug:
            typer.echo("DEBUG: Full name = %s" % bpj.name)
    else:
        typer.secho(
            "Error: You must either provide model or one of --all or --state.", fg="red"
        )


def bulk_run(
    project: str = typer.Option(None, help="Project ID"),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    dest_type: PredictionDestType = typer.Option(
        PredictionDestType.store, help="Destination type.", hidden=True
    ),
    dest_path: Path = typer.Option(
        None, help="Destination path (CSV only)", hidden=True
    ),
    all: bool = typer.Option(False, "--all", help="Run all batch prediction jobs?"),
    state: BatchPredictionState = typer.Option(
        None, help="Only re-run batch predictions job with a certain state."
    ),
    include_estimate: bool = typer.Option(
        False, "--include-estimate", help="Include credit usage estimate for resources."
    ),
):
    """Bulk run batch prediction jobs."""
    if not (all or state is not None):
        typer.secho(
            "Error: You must specify either --all or --state <state> to specify which batch prediction jobs to run. ",
            fg="red",
        )
        raise typer.Exit(code=0)
    else:
        project, environment = utils.get_project_and_environment(project, environment)
        c = Client(project=utils.get_environment_name(project, environment))

        model_list = []

        if all == True:
            model_list = c.models.list_all().to_list()
        else:
            for m in c.models.list_all():
                try:
                    bpj = c.batch_prediction_jobs.get(m.latest_batch_prediction)
                    if bpj.state.value == state:
                        model_list.append(m)
                except:
                    continue  # if no batch prediction has been run yet, this will error, just skip and keep going

        if len(model_list) > 0:
            credits = {}
            endpoint = utils.get_endpoint()
            p_slug = utils.get_environment_name(project, environment).split("/")[-1]
            for model in model_list:
                try:
                    bpj = model.batch_predict(dest_type=dest_type)
                    typer.secho(
                        "Started running batch prediction job %s for model %s in project %s, environment %s.\n"
                        % (bpj.id, model.id, project, environment),
                        fg="green",
                    )
                    typer.secho("You can access the batch prediction job at: ")
                    typer.secho(
                        f"  {endpoint}/projects/{p_slug}/models/{model.id}/batchPredictions/{bpj.id}\n",
                        fg="blue",
                    )
                    if dest_type == PredictionDestType.store:
                        typer.secho("Predictions can be found in feature store at: ")
                        typer.secho(
                            "   %s\n" % (bpj.table),
                            fg="blue",
                        )
                    if include_estimate:
                        credits_used = utils.get_batch_prediction_credit_estimate(
                            model.name, project, environment
                        )
                        credits["%s %s" % (model.id, "PREDICT")] = credits_used
                except Exception as e:
                    typer.secho(
                        "Failed to start batch prediction for model %s in project %s, environment %s: %s"
                        % (model.id, project, environment, str(e)),
                        fg="red",
                    )
                    continue

            if include_estimate and len(credits) > 0:
                utils.print_credit_estimate_summary(credits)

        else:
            typer.secho("No batch prediction found to run.")


@app.command("cancel")
@utils.exit_on_error
def cancel(
    batch_prediction: str = typer.Argument(..., help="Batch prediction ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force cancellation. Skips confirmation."
    ),
):
    """Cancel a batch prediction that is currently running."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    try:
        bpj = c.batch_prediction_jobs.get(batch_prediction)
    except Exception as e:
        typer.secho(
            "Failed to find batch prediction job %s in project %s, environment %s. Please double check that your job id, project id, and environment id are correct. : %s"
            % (batch_prediction, project, environment, str(e)),
            fg="red",
        )
        raise typer.Exit(code=1)
    if not force:
        force = typer.confirm(
            "Are you sure you want to cancel the batch prediction job %s?"
            % batch_prediction
        )
    if force:
        bpj.cancel()
    typer.secho("Successfully cancelled batch prediction %s" % (bpj.id), fg="green")


##dev-2222: removing because we no longer store csvs.
# We may later decide we want download funciton from the cli (might be nice for automation purposes), so leaving in
# @app.command("download")
# @utils.exit_on_error
# def download(
#    batch_prediction: str = typer.Argument(..., help="Batch prediction ID."),
#    project: str = typer.Option(None, help="Project ID."),
#    path: Path = typer.Option(..., help="Local path to save redictions."),
# ):
#    """Download a batch prediction job to CSV."""
#    c = Client(project=project)
#    if project is None:
#        project = utils.get_default_project()
#    bpj = c.batch_prediction_jobs.get(batch_prediction)
#    bpj.download("%s/%s_predictions.csv" % (path, bpj.id))
#    typer.secho(
#        "Successfully downloaded batch prediction job %s to path %s/%s_predictions.csv."
#        % (bpj.id, path, bpj.id),
#        fg="green",
#   )



import typer
import yaml

from continual import continual as c
from continual.python.cli import utils
from itertools import groupby

YAML_TEMPLATE = """type: {{TYPE}}
name: {{NAME}}{{ENTITY}}{{TARGET}}{{INDEX}}{{TIME_INDEX}}{{SPLIT}}{{COLUMNS}}{{EXCLUDED_COLUMNS}}{{TRAINING_CONFIG}}{{PREDICTION_CONFIG}}{{PROMOTION_CONFIG}}{{TABLE_OR_QUERY}}
"""

SQL_STUB_TEMPLATE = """{{
  config(
    materialized = 'view',
    meta = {'continual.enabled' : False},
    )
}}
SELECT * from {{PRED_DB}}
"""

EXPOSURE_TEMPLATE = """
  - name: {{NAME}}
    type: ml
    url: {{URL}}
    description: >
      Created automatically via Continual dbt integration. View more details about this model and its predictions at {{URL}}
    tags: ['continual'] 

    depends_on:{{TABLE_LIST}}
    owner:
      email: {{EMAIL}}
"""

SOURCES_TEMPLATE = """version: 2 

sources: 
    - name: {{SCHEMA}}
      database: {{DATABASE}}
      description: Created by Continual dbt Integration. 
      tags: ['continual']
      tables: {{TABLES}}   
"""

# builds yaml from template
def build_yaml(
    table_name,
    type,
    name,
    entity,
    index,
    time_index,
    target,
    column_config,
    split,
    training_config,
    prediction_config,
    promotion_config,
    excluded_columns,
):
    index_text = ""
    if index is not None:
        index_text = "\nindex: %s" % index

    time_index_text = ""
    if time_index is not None:
        time_index_text = "\ntime_index: %s" % time_index

    entity_text = ""
    if entity is not None:
        entity_text = "\nentity: %s" % entity

    target_text = ""
    if target is not None:
        target_text = "\ntarget: %s" % target

    split_text = ""
    if split is not None:
        split_text = "\nsplit: %s" % split

    training_text = ""
    if training_config is not None:
        training_text = "\ntrain:"
        for t in training_config:
            training_text = (
                training_text
                + """
    %s: %s
    """
                % (t, training_config[t])
            )

    prediction_text = ""
    if prediction_config is not None:
        prediction_text = "\npredict:"
        for p in prediction_config:
            prediction_text = (
                prediction_text
                + """
    %s: %s
    """
                % (p, prediction_config[p])
            )

    promotion_text = ""
    if promotion_config is not None:
        promotion_text = "\npromote:"
        for p in promotion_config:
            promotion_text = (
                promotion_text
                + """
    %s: %s
    """
                % (p, promotion_config[p])
            )

    column_text = ""
    if column_config is not None:
        column_text = "\ncolumns:"
        for col in column_config:
            column_text = (
                column_text
                + """
    - """
            )
            for field in col:
                column_text = (
                    column_text
                    + """%s: %s
      """
                    % (field, col[field])
                )

    excluded_text = ""
    if excluded_columns is not None:
        excluded_text = "\nexclude_columns:"
        for col in excluded_columns:
            excluded_text = (
                excluded_text
                + """
    - %s"""
                % (col)
            )

    source = "table"
    # if type == "Model":
    #    source = "query"
    #    table_name = "select * from %s" % table_name

    # need to properly modify spacing in sql:
    yaml = (
        YAML_TEMPLATE.replace("{{TYPE}}", type)
        .replace("{{NAME}}", name)
        .replace("{{ENTITY}}", entity_text)
        .replace("{{TARGET}}", target_text)
        .replace("{{INDEX}}", index_text)
        .replace("{{TIME_INDEX}}", time_index_text)
        .replace("{{SPLIT}}", split_text)
        .replace("{{TRAINING_CONFIG}}", training_text)
        .replace("{{PREDICTION_CONFIG}}", prediction_text)
        .replace("{{PROMOTION_CONFIG}}", promotion_text)
        .replace("{{COLUMNS}}", column_text)
        .replace("{{EXCLUDED_COLUMNS}}", excluded_text)
        .replace("{{TABLE_OR_QUERY}}", "\n%s: %s" % (source, table_name))
    )
    return yaml


def generate_exposure_yaml(model, table_list, project_slug):
    url = "%s/%s/models/%s" % (utils.get_endpoint(), project_slug, model)

    table_list_text = ""
    for t in table_list:
        table_list_text = (
            table_list_text
            + """
      - ref(\'%s\')"""
            % t
        )

    email = c.config.email

    yaml = (
        EXPOSURE_TEMPLATE.replace("{{NAME}}", model)
        .replace("{{URL}}", url)
        .replace("{{TABLE_LIST}}", table_list_text)
        .replace("{{EMAIL}}", email)
    )
    return yaml


def generate_source_yaml(tables, schema, database):
    table_list_text = ""
    for t in tables:
        table_list_text = (
            table_list_text
            + """
        - name: %s"""
            % t
        )
    yaml = (
        SOURCES_TEMPLATE.replace("{{DATABASE}}", database)
        .replace("{{SCHEMA}}", schema)
        .replace("{{TABLES}}", table_list_text)
    )
    return yaml


def generate_exposures(entity_dict, model_dict, project):
    exposures = []
    for model, entity_list in model_dict.items():
        table_list = []
        if model[-8:] != "__path__":
            for entity in entity_list:
                table_list.extend(entity_dict.get(entity))
            exposure_text = generate_exposure_yaml(model, table_list, project)
            exposures.append((exposure_text, model_dict["%s__path__" % model]))
    return exposures


def generate_sources(sources):
    # we know all model tables in an env have to have the same db and schema
    database = None
    schema = None
    tables = None

    if sources:
        pieces = sources[0].split(".")
        schema = pieces[-2]
        if len(pieces) > 2:  # bigquery doesn't have a db
            database = pieces[0]
        tables = [t.split(".")[-1] for t in sources]

    source_text = generate_source_yaml(tables, schema, database)
    return source_text


# takes continual config and return yaml text. Only used if continual.enabled is not False
def generate_yaml(continual_configs, name, table_name):
    type = get_continual_config(continual_configs, "type")
    name = get_continual_config(
        continual_configs, "name", name
    )  # allow config overwrite of name
    entity = get_continual_config(continual_configs, "entity")
    index = get_continual_config(continual_configs, "index")
    time_index = get_continual_config(continual_configs, "time_index")
    split = get_continual_config(continual_configs, "split")
    target = get_continual_config(continual_configs, "target")
    columns = get_continual_config(continual_configs, "columns")
    excluded_columns = get_continual_config(continual_configs, "exclude_columns")
    training_config = get_continual_config(continual_configs, "train")
    prediction_config = get_continual_config(continual_configs, "prediction")
    promotion_config = get_continual_config(continual_configs, "promotion")
    yaml_text = None
    try:
        if (type is None) or (name is None) or (index is None):
            raise Exception(
                "Insufficient inputs. Required configurations missing: continual_name: %s, continual_type: %s, continual_index: %s"
                % (name, type, index)
            )
        if type.lower() == "model":
            if target is None:
                raise Exception(
                    "Insufficient inputs. Model requires a target but none is provided."
                )
            else:
                yaml_text = build_yaml(
                    table_name,
                    "Model",
                    name,
                    entity,
                    index,
                    time_index,
                    target,
                    columns,
                    split,
                    training_config,
                    prediction_config,
                    promotion_config,
                    excluded_columns,
                )
        elif type.lower() == "featureset":
            yaml_text = build_yaml(
                table_name,
                "FeatureSet",
                name,
                entity,
                index,
                time_index,
                target,
                columns,
                split,
                training_config,
                prediction_config,
                promotion_config,
                excluded_columns,
            )
        else:
            typer.secho(
                "continual_type (%s) unrecognized. Must be Either 'Model' or 'FeatureSet'."
                % type,
                fg="red",
            )
            raise Exception("Unsupported Continual Type.")

    except Exception as e:
        typer.secho(
            "\t Failed to build yaml for dbt model %s: %s" % (name, str(e)), fg="red"
        )

    return yaml_text


# generates model stub
def generate_model_stub(pred_db, create_sources):
    if create_sources:
        pieces = pred_db.split(".")
        table = pieces[-1]
        schema = pieces[-2]
        source_fragment = """{{ source('%s', '%s') }}""" % (schema, table)
        stub_sql = SQL_STUB_TEMPLATE.replace("{{PRED_DB}}", source_fragment)
    else:  # not using sources, so just do a normal query
        stub_sql = SQL_STUB_TEMPLATE.replace("{{PRED_DB}}", pred_db)
    return stub_sql


# stubs are a list of (path, text) tuples
def save_stubs(stubs, print_it=True):
    for (path, text) in stubs:
        with open(path, "w+") as f:
            if print_it:
                typer.secho("Saving stub file %s" % path)
            f.write(text)


# exposures is a list of yaml text. exposures can all live in the same yaml file so we'll add them
def save_exposures(exposures, print_it=True):

    grouped_dict = {}
    for key, val in groupby(
        sorted(exposures, key=lambda ele: ele[1]), key=lambda ele: ele[1]
    ):
        grouped_dict[key] = [ele[0] for ele in val]

    exposure_text = """version: 2
exposures:"""
    for path, text_list in grouped_dict.items():
        for text in text_list:
            exposure_text = "%s\n%s" % (exposure_text, text)
        e_path = "%s/continual_exposures.yml" % path
        with open(e_path, "w+") as f:
            if print_it:
                typer.secho("Saving exposure file %s" % e_path)
            f.write(exposure_text)


def save_sources(sources_text, path, print_it=True):
    s_path = "%s/continual_sources.yml" % path
    with open(s_path, "w+") as f:
        if print_it:
            typer.secho("Saving sources file %s!" % s_path)
        f.write(sources_text)


# processes nodes & generate yaml files
def process_node(
    node, dbt_project_dir, pred_db_prefix, entity_dict, model_dict, sources_list
):
    # if we got to this point, we know we have at least continual.type, so continual_config has to be non-null
    keys = node.get("meta").copy()
    keys.update(
        node.get("config").get("meta")
    )  # model-level config should override project-level config
    continual_config = {k: v for k, v in keys.items() if k.startswith("continual")}
    table_name = node.get("relation_name")
    name = get_continual_config(continual_config, "name", node.get("name"))

    ### do all the yaml processing from dbt-continual plugin
    ### only process node if continual is not disabled
    continual_enabled = get_continual_config(
        continual_config, "enabled", True
    )  # defaults to true
    if continual_enabled:
        # deal with yamls
        try:
            yaml_text = generate_yaml(continual_config, name, table_name)

            # deal with stubs
            stub_path = None
            stub_text = None
            if get_continual_config(continual_config, "type").lower() == "model":
                create_sources = get_continual_config(
                    continual_config, "create_sources"
                )

                pred_db = "%s.model_%s_predictions" % (pred_db_prefix, name)

                if create_sources:
                    sources_list.append(pred_db)

                if get_continual_config(continual_config, "create_stub"):
                    stub_text = generate_model_stub(pred_db, create_sources)
                    stub_path = "%s/%s" % (
                        dbt_project_dir,
                        node.get("original_file_path"),
                    )
                    stub_path = stub_path.replace(".sql", "_predictions.sql")

            # deal with entity + model dicts
            if get_continual_config(continual_config, "create_exposures") == True:
                entity = get_continual_config(continual_config, "entity", name)
                table_list = entity_dict.get(entity, [])
                table_list.append(node.get("name"))
                entity_dict[entity] = table_list
                if get_continual_config(continual_config, "type").lower() == "model":
                    entity_list = [entity]
                    for col in get_continual_config(continual_config, "columns", []):
                        linked_entity = col.get("entity")
                        if linked_entity is not None:
                            entity_list.append(linked_entity)
                    model_dict[name] = entity_list
                    model_dict["%s__path__" % name] = "%s/%s" % (
                        dbt_project_dir,
                        node.get("original_file_path").split("/")[0],
                    )

            return (
                (name, yaml_text),
                (stub_path, stub_text),
                entity_dict,
                model_dict,
                sources_list,
            )

        except Exception as e:
            typer.secho(
                "Failed to generate yaml for model %s: %s" % (name, str(e)), fg="red"
            )
            return None

    else:
        return None


def get_continual_config(config, key, default=None):
    value = None
    if config.get("continual", None):
        value = config.get("continual").get(key, None)
    if value is None:
        value = config.get("continual.%s" % key, None)
    if value is None:
        value = default
    return value


# proceses manifest file, builds yaml files and saves them to target dir
def process_manifest(
    manifest_text, dbt_project_dir, yaml_dir, pred_db_prefix, project, verbose=False
):
    nodes = manifest_text.get("nodes")
    paths = []
    stubs = []
    exposures = []
    sources = []
    entity_dict = {}
    model_dict = {}
    sources_list = []
    if nodes:
        # we only care about model nodes
        model_nodes = {
            k: v for (k, v) in nodes.items() if v.get("resource_type") == "model"
        }
        if model_nodes:
            # must have a continual.type (featureset or model) set or we can't do anything with it
            continual_nodes = {
                k: v
                for (k, v) in model_nodes.items()
                if get_continual_config(v.get("config").get("meta", {}), "type")
                is not None
            }  # this is now safe. dbt version < 0.21 doesn't incldue meta by default, so need to return {} in these instances
            yamls = []
            for k, v in continual_nodes.items():
                # create continual yaml
                (
                    continual_yaml,
                    stub,
                    entity_dict,
                    model_dict,
                    sources_list,
                ) = process_node(
                    v,
                    dbt_project_dir,
                    pred_db_prefix,
                    entity_dict,
                    model_dict,
                    sources_list,
                )
                if continual_yaml:
                    yamls.append(continual_yaml)
                if stub[0] and stub[1]:
                    stubs.append(stub)
            if len(yamls) > 0:
                # save yamls to dbt target dir
                target_dir = "%s/%s" % (dbt_project_dir, yaml_dir)
                paths = utils.save_yamls(yamls, target_dir, verbose)
            if len(model_dict) > 0:
                exposures = generate_exposures(entity_dict, model_dict, project)
            if len(sources_list) > 0:
                sources = generate_sources(sources_list)
    return (paths, stubs, exposures, sources)


def get_datastore(profiles_file, dbt_profile, dbt_target):
    # get profile connetion info
    datastore = None
    try:
        with open(profiles_file, "r") as file:
            text = yaml.load(file, Loader=yaml.FullLoader)
            target_config = text.get(dbt_profile).get("outputs").get(dbt_target)
            type = target_config.get("type")

            # consider not updating user/pass
            if type.lower() == "snowflake":
                config = {
                    # "host" : target_config.get("account"),
                    # "username" : target_config.get("user"),
                    # "password" : target_config.get("password"),
                    "database": target_config.get("database"),
                    "db_schema": target_config.get("schema"),
                    # "warehouse" : target_config.get("warehouse"),
                    # "role" : target_config.get("role",None),
                }
                datastore = {"type": type, "snowflake": config}
            elif type.lower() == "redshift":
                config = {
                    # "host" : target_config.get("host"),
                    # "username" : target_config.get("user"),
                    # "password" : target_config.get("password"),
                    "database": target_config.get("dbname"),
                    "db_schema": target_config.get("schema"),
                    # "port" : target_config.get("port"),
                }
                datastore = {"type": type, "redshift": config}
            elif type.lower() == "bigquery":
                config = {
                    # "auth_file_name" : target_config.get("project"),
                    "dataset": target_config.get("dataset")
                    or target_config.get("schema"),
                    # "auth_file" : target_config.get("keyfile_json"),
                }
                datastore = {"type": type, "big_query": config}
            else:  # not supported
                typer.secho(
                    "DataStore type '%s' is not supported by Continual at this time."
                    % type,
                    fg="red",
                )
    except:
        pass
    return datastore


def get_default_target(profiles_file, dbt_profile):
    dbt_target = None
    try:
        with open(profiles_file, "r") as file:
            typer.secho(f"Reading dbt profile '{dbt_profile}'...", fg="blue")
            text = yaml.load(file, Loader=yaml.FullLoader)
            dbt_target = text.get(dbt_profile).get("target")
    except:
        typer.secho(
            "No default target found for profile '%s' in profiles file %s. Ensure your profile name is valid and either add a default target for your profile or pass in a target via --target"
            % (dbt_profile, profiles_file),
            fg="red",
        )
        exit(1)

    return dbt_target


# get or create environment
def get_or_create_env(client, profiles_file, dbt_target, dbt_profile):
    # get or create env
    env = None
    if dbt_target is not None:
        try:
            env = client.environments.get(dbt_target)
        # doesn't exist, so we'll create it
        except:
            try:
                data_store = get_datastore(profiles_file, dbt_profile, dbt_target)
                env = client.environments.create(
                    dbt_target, data_store=data_store, source=client.config.project
                )
                typer.secho(
                    "Created Continual environment '%s' to match dbt target '%s'.\n"
                    % (env.name, dbt_target),
                    fg="green",
                )
            except Exception as e:
                typer.secho(
                    "Failed to create environment for dbt target '%s': %s"
                    % (dbt_target, str(e)),
                    fg="red",
                )
                exit(1)

    return env.name  # should never be None


def get_environment_datastore_type(env):
    type = env.data_store.type
    if type is None or len(type) == 0:
        type = c.projects.get(env.id.split("@")[0]).data_store.type
    return type


def get_pred_db_prefix(e):
    type = get_environment_datastore_type(e)
    if type.lower() == "snowflake":
        return "%s.%s" % (
            e.data_store.snowflake.database,
            e.data_store.snowflake.db_schema,
        )
    elif type.lower() == "redshift":
        return "%s.%s" % (
            e.data_store.redshift.database,
            e.data_store.redshift.db_schema,
        )
    elif type.lower() == "bigquery":
        return "%s" % (e.data_store.big_query.dataset)
    else:  # not supported
        typer.secho(
            "Datastore type %s is not supported by continual at this time." % type,
            fg="red",
        )
        return None



import typer

from continual.python.cli import utils
from continual import Client

from rich.console import Console

app = typer.Typer(help="Manage organizations.")


def format_organization_data(o, zipped=False):
    create_time = o.create_time.replace(microsecond=0)
    update_time = o.update_time.replace(microsecond=0)
    num_projects = len(o.projects.list(1000))
    data = [o.id, o.display_name, num_projects, create_time, update_time]
    headers = ["ID", "Display Name", "Projects", "Created", "Updated"]

    if zipped:
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(n=30, style=None)


@app.command("list")
@utils.exit_on_error
def list(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List organizations."""
    c = Client()
    data = []
    headers = []
    for org in c.organizations.list(n):
        (o_data, headers) = format_organization_data(org)
        data.append(o_data)
    typer.secho("\nFound %s organizations:" % len(data), fg="blue")
    if style is None:
        style = c.config.style
    style = utils.get_style(style)
    utils.print_table(data, headers, style=style)


@app.command("get")
@utils.exit_on_error
def get(
    organization: str = typer.Argument(..., help="Organization ID."),
    json: bool = typer.Option(False, "--json", help="Print full JSON representation."),
):
    """Get orgaanization details."""
    c = Client()
    try:
        org = c.organizations.get(organization)
    except:
        org = utils.get_org_from_display_name(organization)
    if json:
        console = Console()
        console.print(org.to_dict())
    else:
        data = format_organization_data(org, zipped=True)
        typer.secho("\nRetrieving organization %s: \n" % (org.id), fg="blue")
        utils.print_info(data)


@app.command("create")
@utils.exit_on_error
def create(
    organization: str = typer.Argument(
        ..., help="Organization name.", show_default=False
    ),
):
    """Create organization."""
    org = utils.get_or_create_org(organization)
    typer.secho(
        "Successfully created organization %s with id %s." % (org.display_name, org.id),
        fg="green",
    )


@app.command("delete")
@utils.exit_on_error
def delete(
    organization: str = typer.Argument(..., help="Organization ID."),
):
    """Delete organization."""
    c = Client()
    org = c.organizations.get(organization)
    org.delete()
    typer.secho("Successfully deleted org %s." % organization, fg="green")


@app.command("update")
@utils.exit_on_error
def update(
    organization: str = typer.Argument(..., help="Organization ID"),
    display_name: str = typer.Option(..., help="Display name."),
):
    """Update Org Name."""
    c = Client()
    org = c.organizations.get(organization)
    org.update(display_name)
    typer.secho("Successfully updated organization %s." % organization, fg="green")



import typer
import os
import shutil
import yaml
import json
import click

from continual.python.sdk.client import Client
from continual.python.sdk.identifiers import ProjectEnvironmentIdentifer
from continual.python.cli import projects as Projects, model_versions as ModelVersions

from pathlib import Path
from typing import List
from enum import Enum

import sys

if sys.version_info < (3, 8):
    from importlib_metadata import version
else:
    from importlib.metadata import version


from continual.python.cli import (
    feature_sets,
    models,
    model_versions,
    experiments,
    projects,
    environments,
    organizations,
    users,
    featurestore_config,
    batch_predictions,
    promotions,
    changes,
    utils,
    events,
    reports,
    config,
    utils_dbt,
)
from continual.python.sdk.exceptions import AlreadyExistsError
from typing import Optional

try:
    __version__ = version("continual")
except:
    __version__ = "local-dev"


def version_callback(value: bool):
    if value:
        typer.echo(f"Continual CLI Version: {__version__}")
        raise typer.Exit()


# ************************#
# Define Typer Structure #
# ************************#


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx):
        return self.commands.keys()


cli = typer.Typer(cls=NaturalOrderGroup, help="Continual – Operational AI. Simplified.")

cli.add_typer(feature_sets.app, name="feature-sets")
cli.add_typer(models.app, name="models")
cli.add_typer(model_versions.app, name="model-versions")
cli.add_typer(experiments.app, name="experiments")
cli.add_typer(promotions.app, name="promotions")
cli.add_typer(batch_predictions.app, name="batch-predictions")
cli.add_typer(projects.app, name="projects")
cli.add_typer(environments.app, name="environments")
cli.add_typer(organizations.app, name="organizations")
cli.add_typer(changes.app, name="changes")
cli.add_typer(users.app, name="users")
# cli.add_typer(ingestions.app, name="ingestions")
cli.add_typer(featurestore_config.app, name="warehouse-config", hidden=True)
cli.add_typer(events.app, name="events")
cli.add_typer(reports.app, name="report", hidden=True)
cli.add_typer(config.app, name="config")


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print Continual CLI version",
    ),
):
    return


class ContinualDemo(str, Enum):
    bankmarketing = "bank_marketing"
    kickstarter = "kickstarter"


class CloudDataWarehouse(str, Enum):
    snowflake = "snowflake"
    redshift = "redshift"
    bigquery = "bigquery"
    # synapse = "synapse"
    # sqlserver = "sqlserver"
    # firebolt = "firebolt"


@cli.command("login")
@utils.exit_on_error
def login(
    email: str = typer.Option(..., prompt=True, help="Email address."),
    password: str = typer.Option(..., prompt=True, hide_input=True, help="Password."),
    project: str = typer.Option(None, help="Default project."),
    endpoint: str = typer.Option(
        None,
        envvar="CONTINUAL_ENDPOINT",
        help="Optional endpoint for private cloud installations.",
    ),
):
    """Log in to Continual.

    Logs in to Continual and saves the session information to
    ~/.continual/continual.yaml which is used by the SDK/CLI. All
    future commands will use the session information to authenticate
    and use the default project if not explicitly passed.
    """
    client = Client(endpoint=endpoint)
    user = client.login(email=email, password=password)
    if project is None:
        client.config.project = None
        projects = client.projects.list_all()
        i = 1
        proj_list = []
        if len(list(projects)) > 0:
            for project in projects:
                if i == 1:
                    typer.echo("Select default project:")
                project_id = project.name.split("/")[-1]
                proj_list.append(project.name)
                typer.echo(f" [{i}]: {project.display_name} ({project_id})")
                i += 1
            key = typer.prompt("Choice")
            client.config.project = proj_list[int(key) - 1]
            client.config.environment = "prod"
        else:
            typer.secho(
                "You have no projects created in continual. You should create a project in the Continual Web UI and then use `continual config set-project <project>` to use the CLI."
            )
    else:
        client.config.project = project
        client.config.environment = "prod"
    client.config.save()
    typer.echo(
        f"🚀 Login successful.  Welcome to Continual, {user.first_name} {user.last_name}."
    )


@cli.command("logout")
@utils.exit_on_error
def logout():
    """Log out of Continual.

    Logs out of the current session.
    """
    c = Client()
    c.logout()
    typer.echo("👋 See you later, come back soon!")


@cli.command("checkout")
@utils.exit_on_error
def checkout(
    environment: str = typer.Argument(
        ...,
        envvar="CONTINUAL_ENVIRONMENT",
        help="Unique environment ID within a project.",
    ),
    project: str = typer.Option(None, envvar="CONTINUAL_PROJECT", help="Project ID."),
    strict: bool = typer.Option(False, help="Auto-format invalid environment IDs."),
    wait: bool = typer.Option(
        False, help="Wait for environment to be synced with the base environment."
    ),
):
    """
    Checkout Continual environment.
    """
    project, environment = utils.get_project_and_environment(project, environment)
    client = Client(project=project)

    project_name = client.config.project
    project_id = project_name.split("/")[-1]

    if not strict:
        if environment[0] == "_":
            environment = environment[1:]
        if environment[-1] == "_":
            environment = environment[:-1]

        formatted = ""
        for ch in list(environment):
            if not ch.isalnum() and formatted[-1] and formatted[-1] != "_":
                formatted += "_"
            elif ch.isalnum():
                formatted += ch.lower()
        environment = formatted

    try:
        env = client.environments.create(id=environment, source=project_name)

        client.config.set_project(project, save=False)
        client.config.set_environment(environment, save=False)
        client.config.save()

        changes = env.changes.list(5)
        if not changes:
            typer.secho(
                f"Successfully created environment ({environment}). It is even with ({project_name}).",
                fg="green",
            )
            exit(0)

        push = changes[0]

        typer.secho(
            f"Creating environment ({environment}) and syncing with ({project_name}):",
            fg="blue",
        )

        utils.print_change(
            push,
            project_id,
            environment,
            f"Creating environment ({environment}): initial sync with ({project_name})",
        )

    except Exception as e:
        if isinstance(e, AlreadyExistsError):
            client.config.set_project(project, save=False)
            client.config.set_environment(environment, save=False)
            client.config.save()

            typer.secho(
                f"Switched active environment for project ({project_id}) to environment ({environment}).",
                fg="green",
            )
            exit(0)
        else:
            typer.secho(
                "Something went wrong while creating the environment. %s" % str(e),
                fg="red",
            )
            exit(1)


@cli.command("merge")
@utils.exit_on_error
def merge(
    from_environment: str = typer.Option(
        ..., "--from", help="Environment to merge changes from."
    ),
    to_environment: str = typer.Option(
        "prod", "--to", help="Environment to merge changes to. Defaults to production."
    ),
    project: str = typer.Option(None, help="Project ID.", envvar="CONTINUAL_PROJECT"),
    message: str = typer.Option(
        None, help="Optional message to provide context for your merge."
    ),
    plan_only: bool = typer.Option(
        False, "--plan-only", help="View the execution plan without applying changes."
    ),
):
    """Merge changes from one environment to another.

    Pushes all Models and Feature Sets from one environment (--from) into another
    environment (--to).
    """

    client = Client()
    project, from_environment_id = utils.get_project_and_environment(
        project, from_environment
    )
    project, to_environment_id = utils.get_project_and_environment(
        project, to_environment
    )

    if from_environment_id and from_environment == to_environment_id:
        typer.secho("Identical environments provided. Nothing to do.", fg="gray")
        exit(0)
    elif not from_environment_id:
        typer.secho("From environment (--from) got no value.", fg="red")
        exit(1)

    env_slug = utils.get_environment_name(project, to_environment_id)
    from_env_slug = utils.get_environment_name(project, from_environment_id)

    env = client.environments.get(env_slug)

    message = utils.get_message(
        message, f"Merge environment {from_environment_id} into {to_environment_id}"
    )

    push = env.changes.push(
        paths=[],
        resource=from_env_slug,
        message=message,
        trigger="CLI",
        plan_only=plan_only,
    )

    utils.print_change(push, project, to_environment_id, message)


def push_paths(
    paths: List[str],
    client: Client,
    message: str = "",
    plan_only: bool = False,
    purge: bool = False,
    force: bool = False,
    trigger: str = "CLI",
    resource: str = "",
    train_all: bool = False,
    train_none: bool = False,
    update_all: bool = False,
    profile_none: bool = False,
):
    return client.changes.push(
        paths=paths,
        plan_only=plan_only,
        purge=purge,
        update_all=force or update_all,
        message=message,
        trigger=trigger,
        resource=resource,
        train_all=train_all,
        train_none=train_none,
        profile_none=profile_none,
    )


def print_model_summary(push, project, environment, comparison_env=None):
    current_env = utils.get_environment_name(project, environment)
    c_current = Client(project=current_env)
    if comparison_env is None:
        comparison_env = current_env
        if current_env[-4:] != "prod":
            comparison_env = utils.get_environment_name(project, "prod")
    else:
        comparison_env = utils.get_environment_name(project, comparison_env)

    models = []
    models_single = []
    for step in push.plan:
        if step.operation.value == "TRAIN":
            resource_name = step.resource_name
            parts = resource_name.split("/")
            if parts[2] == "models":
                model_id = parts[-1]
                try:
                    latest_model_version = c_current.models.get(
                        model_id
                    ).latest_model_version
                except Exception:
                    typer.secho("Unable to get model %s" % model_id, fg="red")
                    continue
                if (
                    current_env != comparison_env
                ):  # envs are different. Comapre latest mv in current env to current mv in comparison_env
                    c_comparison = Client(project=comparison_env)
                    try:
                        comparison_model_version = c_comparison.models.get(
                            model_id
                        ).current_version
                        models.append((comparison_model_version, latest_model_version))
                    except Exception:
                        typer.secho(
                            "Unable to find comparison model version for model %s in env %s."
                            % (model_id, comparison_env),
                            fg="blue",
                        )
                        models_single.append(latest_model_version)
                else:  # envs are the same. compare compare latest mv to most recent promoted mv that wasn't the same mv
                    comparison_model_version = c_current.models.get(
                        model_id
                    ).current_version
                    if (
                        latest_model_version == comparison_model_version
                    ):  # If latest = current, then we have to get the promoted mv before this
                        promoted_mvs = sorted(
                            c_current.promotions.list_all(),
                            key=lambda x: x.promoted_time,
                            reverse=True,
                        )
                        if (
                            len(promoted_mvs) > 1
                        ):  # check to make sure we're just not the first time this model has been trained
                            comparison_model_version = promoted_mvs[1].model_version
                            models.append(
                                (comparison_model_version, latest_model_version)
                            )
                        else:
                            models_single.append(latest_model_version)
                    else:
                        models.append((comparison_model_version, latest_model_version))

    for model in set(models):
        typer.secho(
            "\nComparing \n\t%s in environment %s [original] to \n\t%s in environment %s [new]..."
            % (model[0], comparison_env, model[1], current_env),
            fg="blue",
        )
        ModelVersions.compare(
            right_model_version=model[0],
            right_project=project,
            right_environment=comparison_env,
            left_model_version=model[1],
            left_project=project,
            left_environment=current_env,
            performance_metric=None,
            split_type=utils.SplitType.TEST,
            stacked=True,
            style=c_current.config.style,
        )

    for model in set(models_single):
        ModelVersions.get(
            model_version=model, project=project, environment=current_env, to_json=False
        )


def push_continual(
    paths: List[str],
    client: Client,
    project: str,
    environment: str = "",
    plan_only: bool = False,
    purge: bool = False,
    force: bool = False,
    message: str = "",
    resource: str = "",
    train_all: bool = False,
    train_none: bool = False,
    update_all: bool = False,
    profile_none: bool = False,
    wait: bool = False,
    comparison_env=None,
    include_estimate: bool = False,
):
    project, environment = utils.get_project_and_environment(project, environment)

    # typer path type is PosixPath, which our SDK doesn't expect. Only works w/ strings or list of strings.
    paths = [str(path) for path in paths]
    if paths:
        typer.secho(
            "Pushing path(s) %s into Continual. Project: %s, Environment: %s "
            % (str(paths), project, environment),
            fg="blue",
        )
    elif resource:
        typer.secho(
            "Pushing resource %s into Continual. Project: %s, Environment %s "
            % (resource, project, environment),
            fg="blue",
        )

    try:
        push = push_paths(
            paths=paths,
            client=client,
            message=message,
            plan_only=plan_only,
            purge=purge,
            force=force,
            resource=resource,
            train_all=train_all,
            train_none=train_none,
            update_all=update_all,
            profile_none=profile_none,
        )
        utils.print_change(push, project, environment, message, include_estimate)
        if wait:
            typer.secho("Waiting for all change steps to complete...", fg="magenta")
            push.wait()
            typer.secho("All steps completed!\n", fg="green")
            c = Client(project=utils.get_environment_name(project, environment))
            push = c.changes.get(push.id)  # have to get push again to get updated info
            utils.print_change_plan(push, project, environment)
            print_model_summary(push, project, environment, comparison_env)
    except Exception as e:
        typer.secho(
            "Failed to push path(s) %s into Continual: %s" % (str(paths), str(e)),
            fg="red",
        )


def push_dbt(
    client: Client,
    project: str,
    dbt_project_dir: str = "",
    plan_only: bool = False,
    purge: bool = False,
    force: bool = False,
    message: str = "",
    dbt_profiles_dir: str = "",
    dbt_target: str = "",
    dbt_profile: str = "",
    continual_dir: str = "",
    resource: str = "",
    train_all: bool = False,
    train_none: bool = False,
    update_all: bool = False,
    profile_none: bool = False,
    verbose: bool = False,
    wait: bool = False,
    comparison_env=None,
    include_estimate: bool = False,
):

    dbt_project = "%s/dbt_project.yml" % dbt_project_dir
    dbt_project_name = ""
    # get manifest file
    try:
        with open(dbt_project, "r") as file:
            typer.secho(f"Reading dbt config {dbt_project}...", fg="blue")
            text = yaml.load(file, Loader=yaml.FullLoader)
            dbt_project_name = text.get("name")
            if not dbt_project_name:
                raise Exception(
                    f"Could not read dbt project name from '{dbt_project}'."
                )
            typer.secho(f"Using dbt project {dbt_project_name}.\n", fg="green")
            dbt_target_dir = text.get("target-path", "target")  # default is target
            if not dbt_profile:
                dbt_profile = text.get("profile")
            typer.secho(f"Reading dbt manifest...", fg="blue")
            manifest_file = "%s/%s/manifest.json" % (dbt_project_dir, dbt_target_dir)
            typer.secho(
                "Successfully read manifest file at %s\n" % (manifest_file), fg="green"
            )
    except Exception as e:
        typer.secho(
            "Failed to open dbt project file: %s. Please re-run the command from a dbt project directory or use --project-dir to pass in the path to the directory: %s"
            % (dbt_project, str(e)),
            fg="red",
        )
        exit(1)

    message = utils.get_message(
        message, "Pushing dbt project %s from the CLI" % dbt_project_name
    )

    # get/create env from dbt target
    try:
        # if no target passed in, use what is the default in the profiles.yml file for the given profile
        # if no default is set, the following errors and exits
        profiles_file = "%s/profiles.yml" % (dbt_profiles_dir)
        if not dbt_target:
            dbt_target = utils_dbt.get_default_target(profiles_file, dbt_profile)
        typer.secho(f"Using dbt target '{dbt_target}'.\n", fg="green")

        typer.secho(
            "Determining Continual environment...",
            fg="blue",
        )
        env_name = utils_dbt.get_or_create_env(
            client, profiles_file, dbt_target, dbt_profile
        )
        # reinstatiate client w/ env
        client = Client(project=env_name)
        env = client.environments.get(env_name)
        project_env_id = ProjectEnvironmentIdentifer(project, env.name)
        if project_env_id.environment_is_default:
            typer.secho(
                "Using default Continual environment '%s'\n"
                % (project_env_id.environment_id_alias),
                fg="green",
            )
        else:
            typer.secho(
                "Using Continual environment '%s'\n"
                % (project_env_id.environment_id_alias),
                fg="green",
            )
    except Exception as e:
        typer.secho(
            "Unable to get or create environment for target '%s' and profile '%s': %s"
            % (dbt_target, dbt_profile, str(e)),
            fg="red",
        )
        exit(1)

    if env.data_store.type == "":
        web_ui_project_envs_url = (
            f"{utils.get_endpoint()}/{project_env_id.url_path}/environments"
        )
        typer.secho(
            f"Continual environment '{project_env_id.environment_id_alias}' does not have a data warehouse configured.",
            fg="red",
        )
        typer.secho(
            f"  Connect your data warehouse using Web UI: {web_ui_project_envs_url}",
            fg="red",
        )
        exit(1)

    # need prediction db prefix for stubs
    pred_db_prefix = utils_dbt.get_pred_db_prefix(env)
    # process manifest
    try:
        typer.secho("Processing dbt manifest...", fg="blue")
        with open(manifest_file, "r") as file:
            text = json.load(file)
            # process manifest, save yamls to dbt target dir, we can then push those
            if continual_dir:
                yaml_dir = continual_dir
            else:
                yaml_dir = "%s/continual" % dbt_target_dir
            paths, stubs, exposures, sources = utils_dbt.process_manifest(
                text, dbt_project_dir, yaml_dir, pred_db_prefix, env_name, verbose
            )
            typer.secho(
                "Successfully processed manifest and generated Continual yamls in %s/%s\n"
                % (dbt_project_dir, yaml_dir),
                fg="green",
            )
    except Exception as e:
        typer.secho(
            "Failed to process manifest file %s: %s" % (manifest_file, str(e)), fg="red"
        )
        exit(1)

    #  we could create project here, but it's a bit awkward for dbt users. Why?
    ## to create a project we'd have to know the organization. And not the name, but the ID.
    ## expecting them to annotate that is probably more of a hassle than just creating a project in the UI.
    ## once project is created we can add env w/o worrying aobut orgs.
    if len(paths) > 0:
        try:
            typer.secho(
                "Pushing path(s) %s into Continual. Project: %s, Environment: %s"
                % (
                    str(paths),
                    project_env_id.project_id,
                    project_env_id.environment_id_alias,
                ),
                fg="blue",
            )
            paths = [str(path) for path in paths]
            push = push_paths(
                paths=paths,
                client=client,
                message=message,
                plan_only=plan_only,
                purge=purge,
                force=force,
                resource=resource,
                train_all=train_all,
                train_none=train_none,
                update_all=update_all,
                profile_none=profile_none,
            )
            utils.print_change(
                push,
                project_env_id.project_id,
                project_env_id.environment_id,
                message,
                include_estimate,
            )
            typer.secho("Successfully pushed all dbt models found!", fg="green")
            if sources:
                typer.secho("\nSources enabled. Saving sources files...", fg="blue")
                path = "%s/%s" % (
                    dbt_project_dir,
                    text.get("source-paths", ["models"])[0],
                )
                utils_dbt.save_sources(sources, path, verbose)
                typer.secho("Successfully saved sources files!", fg="green")
            if stubs:
                typer.secho("\nStubs enabled. Saving stub files...", fg="blue")
                utils_dbt.save_stubs(stubs, verbose)
                typer.secho("Successfully saved stub files!", fg="green")
            if exposures:
                typer.secho(
                    "\nExposures enabled. Saving exposure yaml files...", fg="blue"
                )
                utils_dbt.save_exposures(exposures, verbose)
                typer.secho("Successfully saved exposure files!", fg="green")
            if wait:
                typer.secho("Waiting for all change steps to complete...", fg="blue")
                push.wait()
                typer.secho("All steps completed!\n", fg="green")
                c = Client(project=project_env_id.environment_name)
                push = c.changes.get(
                    push.id
                )  # have to get push again to get updated info
                utils.print_change_plan(
                    push, project_env_id.project_id, project_env_id.environment_id
                )
                print_model_summary(
                    push,
                    project_env_id.project_id,
                    project_env_id.environment_id,
                    comparison_env,
                )
        except Exception as e:
            typer.secho("Failed to push paths %s: %s" % (str(paths, str(e))), fg="red")

    else:
        typer.secho(
            "No dbt models found with appropriate continual configuration. Make sure models are annotated with meta tags appropriately and continual is not disabled on the project level.",
            fg="red",
        )


def push_quickstart(
    paths: List[str],
    client: Client,
    project: str = "",
    environment: str = "",
    plan_only: bool = False,
    purge: bool = False,
    force: bool = False,
    message: str = "",
    resource: str = "",
    train_all: bool = False,
    train_none: bool = False,
    update_all: bool = False,
    profile_none: bool = False,
):
    project, environment = utils.get_project_and_environment(project, environment)

    local_project = os.path.join(os.getcwd(), project)
    if os.path.exists(local_project) or resource:
        paths = [local_project]
        push = push_paths(
            paths=paths,
            client=client,
            message=message,
            plan_only=plan_only,
            purge=purge,
            force=force,
            resource=resource,
            train_all=train_all,
            train_none=train_none,
            update_all=update_all,
            profile_none=profile_none,
        )
        utils.print_change(push, project, environment, message)
    else:
        typer.secho(
            "No paths detected and no local project found %s. Please try `continual push` again while providing path arguments."
            % str(local_project),
            fg="red",
        )


@cli.command("push")
@utils.exit_on_error
def push(
    paths: List[Path] = typer.Argument(
        None, exists=True, dir_okay=True, help="Path to YAML files or directories."
    ),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment Name"),
    plan_only: bool = typer.Option(
        False, "--plan-only", help="Display plan only. Do not execute."
    ),
    purge: bool = typer.Option(
        False,
        "--purge",
        help="Delete existing resources not found in pushed configuration.",
    ),
    force: bool = typer.Option(
        False, "--force", help="Force update. Same as --update-all."
    ),
    message: str = typer.Option("", help="Push message."),
    resource: str = typer.Option(
        "", help="Project, Environment, Feature Set, or Model resource name to push."
    ),
    train_all: bool = typer.Option(
        False, "--train-all", help="Train all models in the project."
    ),
    train_none: bool = typer.Option(
        False,
        "--train-none",
        "--no-train",
        help="Do not train Models. NOTE: --train-all will override this option.",
    ),
    update_all: bool = typer.Option(False, "--update-all", help="Force update."),
    profile_none: bool = typer.Option(
        False,
        "--profile-none",
        "--no-profile",
        help="Do not profile resources. NOTE: --train-all will override this option.",
    ),
    dbt_project_dir: Path = typer.Option(
        str(os.environ.get("CONTINUAL_DBT_PROJECT_DIR_DEFAULT", Path.cwd())),
        "--project-dir",
        help="dbt project directory",
        hidden=True,
    ),
    dbt_profiles_dir: Path = typer.Option(
        str(
            os.environ.get(
                "CONTINUAL_DBT_PROFILES_DIR_DEFAULT", os.path.join(Path.home(), ".dbt")
            )
        ),
        "--profiles-dir",
        help="dbt profiles directory",
        hidden=True,
    ),
    dbt_profile: str = typer.Option(
        None,
        "--profile",
        help="dbt profile. Overides profile found in dbt_project.yml",
        hidden=True,
    ),
    dbt_target: str = typer.Option(
        None,
        "--target",
        help="dbt target. Overrides target found in profiles.yml.",
        hidden=True,
    ),
    continual_dir: str = typer.Option(
        None,
        help="directory within dbt project to save continual yaml files. Defaults to ./<target-path>/continual/",
        hidden=True,
    ),
    wait: bool = typer.Option(
        False, "--wait", help="Wait for all steps in the generated plan to finish."
    ),
    comparison_env: str = typer.Option(
        None,
        help="Environment to compare all results with. Only applicable if using '--wait'.",
    ),
    include_estimate: bool = typer.Option(
        False,
        "--include-estimate",
        help="Include credit usage estimate for resources and change.",
    ),
):
    """Push changes to project.

    Pushes all feature set definitions to the project and executes
    the resulting change plan.  The PATHS can be a list of specific feature sets
    or folders of feature set definitions.

    By default, any changed feature sets are reingested and models are
    retrained and promoted.  No feature sets or models are deleted unless
    --purge is passsed. For fine-grained control over the created plan,
    pass additional flags to your push command.
    """

    project, environment = utils.get_project_and_environment(project, environment)
    client = Client(project="%s@%s" % (project, environment))

    message = utils.get_message(
        message, "Pushing path(s) %s from CLI." % str([path.name for path in paths])
    )

    if len(paths) > 0 or resource:  # do normal continual stuff
        push_continual(
            paths=paths,
            client=client,
            project=project,
            environment=environment,
            plan_only=plan_only,
            purge=purge,
            force=force,
            message=message,
            resource=resource,
            train_all=train_all,
            train_none=train_none,
            update_all=update_all,
            profile_none=profile_none,
            wait=wait,
            comparison_env=comparison_env,
            include_estimate=include_estimate,
        )
    elif os.path.exists(
        "%s/dbt_project.yml" % dbt_project_dir
    ):  # we are using dbt project, do dbt integration
        push_dbt(
            client=client,
            project=project,
            dbt_project_dir=dbt_project_dir,
            plan_only=plan_only,
            purge=purge,
            force=force,
            message=message,
            dbt_profiles_dir=dbt_profiles_dir,
            dbt_target=dbt_target,
            dbt_profile=dbt_profile,
            continual_dir=continual_dir,
            resource=resource,
            train_all=train_all,
            train_none=train_none,
            update_all=update_all,
            profile_none=profile_none,
            wait=wait,
            comparison_env=comparison_env,
            include_estimate=include_estimate,
        )
    else:  # try to use quickstart hack
        push_quickstart(
            paths,
            client,
            project,
            environment,
            plan_only,
            purge,
            force,
            message,
            resource,
            train_all,
            train_none,
            update_all,
            profile_none,
        )


@cli.command("run")
@utils.exit_on_error
def run(
    project: str = typer.Option("", help="Continual Project ID."),
    plan_only: bool = typer.Option(
        False, "--plan-only", help="Display plan only. Do not execute."
    ),
    purge: bool = typer.Option(
        False,
        "--purge",
        help="Delete existing resources not found in pushed configuration.",
    ),
    force: bool = typer.Option(False, "--force", help="Force update"),
    message: str = typer.Option("", help="Push message."),
    dbt_project_dir: Path = typer.Option(
        str(os.environ.get("CONTINUAL_DBT_PROJECT_DIR_DEFAULT", Path.cwd())),
        "--project-dir",
        help="dbt project directory",
    ),
    dbt_profiles_dir: Path = typer.Option(
        str(
            os.environ.get(
                "CONTINUAL_DBT_PROFILES_DIR_DEFAULT", os.path.join(Path.home(), ".dbt")
            )
        ),
        "--profiles-dir",
        help="dbt profiles directory",
    ),
    dbt_profile: str = typer.Option(
        None, "--profile", help="dbt profile. Overides profile found in dbt_project.yml"
    ),
    dbt_target: str = typer.Option(
        None, "--target", help="dbt target. Overrides target found in profiles.yml."
    ),
    continual_dir: str = typer.Option(
        None,
        help="directory within dbt project to save continual yaml files. Defaults to ./<target-path>/continual/",
    ),
    resource: str = typer.Option(
        "", help="Project, Environment, Feature Set, or Model resource name to push."
    ),
    train_all: bool = typer.Option(
        False, "--train-all", help="Train all models in the project."
    ),
    train_none: bool = typer.Option(
        False,
        "--train-none",
        "--no-train",
        help="Do not train Models. NOTE: --train-all will override this option.",
    ),
    update_all: bool = typer.Option(False, "--update-all", help="Force update."),
    profile_none: bool = typer.Option(
        False,
        "--profile-none",
        "--no-profile",
        help="Do not profile resources. NOTE: --train-all will override this option.",
    ),
    verbose: bool = typer.Option(
        False, "--verbose", help="Print additional info during run."
    ),
    include_estimate: bool = typer.Option(
        False,
        "--include-estimate",
        help="Include credit usage estimate for resources and change.",
    ),
):
    """Execute Continual on a dbt project.

    Constructs Continual yaml files from dbt models and pushes features and models into Continual.

    Several dbt parameters are supported: --profile, --target, --project-dir, --profiles-dir
    """

    project = utils.get_project(project)
    # skip environment processing, will infer from dbt target
    client = Client(project=project)

    push_dbt(
        client,
        project,
        dbt_project_dir,
        plan_only,
        purge,
        force,
        message,
        dbt_profiles_dir,
        dbt_target,
        dbt_profile,
        continual_dir,
        resource,
        train_all,
        train_none,
        profile_none,
        update_all,
        verbose,
        include_estimate=include_estimate,
    )


@cli.command("quickstart")
@utils.exit_on_error
def quickstart(
    email: str = typer.Option(..., prompt=True, help="Email address."),
    password: str = typer.Option(..., prompt=True, hide_input=True, help="Password."),
    project: str = typer.Option(None, help="Project ID."),
    demo: ContinualDemo = typer.Option("bank_marketing", help="Demo to create."),
    endpoint: str = typer.Option(
        None,
        envvar="CONTINUAL_ENDPOINT",
        help="Optional endpoint for private cloud installations.",
    ),
    create_project: bool = typer.Option(
        False,
        help="Create Project in Continual? Use only if project does not already exist.",
    ),
    org: str = typer.Option(None, help="Org Name. Use only if creating a project"),
    feature_store: str = typer.Option(
        None, help="Feature Store config name. Use only if creating a project"
    ),
):
    """Initializes a Continual demo project.

    This quick start will:

    1. Log you into Continual.

    2. Create project folder locally with sample yaml.

    3. Optionally creates project in Continual.

    4. Seed demo CSVs into project data warehouse.

    This is meant for demonstration purposes and is not recommended for production use cases.
    """
    login(email=email, password=password, project=project, endpoint=endpoint)
    project_id = utils.get_project(project)
    c = Client(project=project_id)

    try:
        c.projects.get(project_id)
    except:
        if create_project:
            Projects.create(project=project_id, org=org, feature_store=feature_store)
        else:
            typer.secho(
                "Error: project %s doesn't exist. You can optionally create it by using the '--create-project' flag. Otherwise, please create the project in the Continual Web UI."
                % project_id,
                fg="red",
            )
            raise typer.Exit(code=1)

    p = c.projects.get(project_id)
    local_path = os.path.join(os.getcwd(), project_id)
    demo_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "examples",
        demo.value,
    )

    if not os.path.exists(local_path):
        os.mkdir(local_path)

    for demo_file in os.listdir(demo_path):
        if ".yaml" in demo_file:
            shutil.copy(os.path.join(demo_path, demo_file), local_path)
        if ".csv" in demo_file:
            typer.secho(
                "Seeding project database with file %s. This may take a few minutes..."
                % demo_file,
                fg="blue",
            )
            (schema, db) = p.seed(os.path.join(demo_path, demo_file))
            typer.secho(
                "Created database table %s in schema %s" % (db, schema), fg="green"
            )

    typer.secho(
        "Created folder %s with %s demo files. You may begin building models by executing: "
        % (local_path, demo.value),
        fg="green",
    )
    typer.secho("\t continual push %s" % local_path, fg="blue")


if __name__ == "__main__":
    cli()



from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.projects import OrganizationProjectManager
from google.protobuf import field_mask_pb2
from continual.python.sdk.events import EventManager


class OrganizationManager(Manager):
    """Manages user resources."""

    name_pattern: str = "organizations/{user}"

    def get(self, id: str) -> Organization:
        """Get organization.

        Arguments:
            id: Organization name or id.

        Returns
            An organization.
        """
        req = management_pb2.GetOrganizationRequest(name=self.name(id))
        resp = self.client._management.GetOrganization(req)
        return Organization.from_proto(resp, client=self.client)

    def list(self, page_size: Optional[int] = None) -> List[Organization]:
        """List organizations.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of organizations.
        """
        req = management_pb2.ListOrganizationsRequest(page_size=page_size)
        resp = self.client._management.ListOrganizations(req)
        return [
            Organization.from_proto(u, client=self.client) for u in resp.organizations
        ]

    def list_all(self) -> Pager[Organization]:
        """List all organizations.

        Pages through all organization using an iterator.

        Returns:
            A iterator of all organizations.
        """

        def next_page(next_page_token):
            req = management_pb2.ListOrganizationsRequest(page_token=next_page_token)
            resp = self.client._management.ListOrganizations(req)
            return (
                [
                    Organization.from_proto(u, client=self.client)
                    for u in resp.organizations
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def create(self, display_name: str) -> Organization:
        """Create an organization.

        Arguments:
            display_name: Display name.

        Returns:
            A new organization.
        """
        req = management_pb2.CreateOrganizationRequest(
            organization=types.Organization(display_name=display_name).to_proto()
        )
        resp = self.client._management.CreateOrganization(req)
        return Organization.from_proto(resp, client=self.client)

    def delete(self, id: str) -> None:
        """Delete an organization.

        Arguments:
            id: Organization name or id.
        """

        req = management_pb2.DeleteOrganizationRequest(name=self.name(id))
        self.client._management.DeleteOrganization(req)

    def update(
        self,
        id: str,
        display_name: Optional[str] = None,
    ) -> Organization:
        """Update organization.

        Arguments:
            display_name:  Display name.

        Returns:
            Updated organization.
        """
        paths = []
        if display_name is not None:
            paths.append("display_name")
        req = management_pb2.UpdateOrganizationRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            organization=Organization(
                name=self.name(id),
                display_name=display_name,
            ).to_proto(),
        )
        resp = self.client._management.UpdateOrganization(req)
        return Organization.from_proto(resp, client=self.client)


class Organization(Resource, types.Organization):
    """Organization resource."""

    manager: OrganizationManager
    projects: OrganizationProjectManager
    events: EventManager
    name_pattern = "organizations/{users}"

    def _init(self):
        self.manager = OrganizationManager(parent=self.parent, client=self.client)
        self.projects = OrganizationProjectManager(parent=self.name, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def delete(self) -> None:
        """Delete organization."""
        self.manager.delete(self.name)

    def update(
        self,
        display_name: Optional[str] = None,
    ) -> Organization:
        """Update organization.

        Arguments:
            display_name:  Display name.

        Returns:
            Updated organization.
        """
        return self.manager.update(
            self.name,
            display_name=display_name,
        )



# fmt: off
# Generated by the protocol buffer compiler.  DO NOT EDIT!
from __future__ import annotations
from datetime import datetime, timedelta
import enum
from typing import Any, Dict, List, Optional, TypeVar, Type
from google.protobuf import timestamp_pb2
from google.protobuf import duration_pb2
from google.protobuf.json_format import MessageToDict

from continual.rpc.graphql import options_pb2


T = TypeVar("T")

def _to_proto_dict(x: dict) -> dict:
    """Converts dict to proto friendly representation."""
    out = dict()
    for key, item in x.items():
        if isinstance(item, list):
            out_list = []
            for e in item:
                if isinstance(e, datetime):
                    dt = timestamp_pb2.Timestamp()
                    dt.FromDatetime(e)
                    out_list.append(dt)
                elif isinstance(e, timedelta):
                    td = duration_pb2.Duration()
                    td.FromTimedelta(e)
                    out_list.append(td)
                elif hasattr(e, "to_proto"):
                    out_list.append(e.to_proto())
                else:
                    out_list.append(e)
            out[key] = out_list
        elif isinstance(item, datetime):
            dt = timestamp_pb2.Timestamp()
            dt.FromDatetime(item)
            out[key] = dt
        elif isinstance(item, timedelta):
            td = duration_pb2.Duration()
            td.FromTimedelta(item)
            out[key] = td
        elif hasattr(item, "to_proto"):
            out[key] = item.to_proto()
        else:
            out[key] = item
    return out


def _set_attribute(
    attributes: dict, class_name: str, field: str, value: Any, fieldType: Any, oneOfField: str
):
    """Helper that sets attribute on the object.
    
    Accepts dicts for classes and normalizes enums.
    """
    if isinstance(value, dict) and getattr(fieldType, "__origin__", None) == dict:
        attributes[field] = value
    elif isinstance(value, dict):
        attributes[field] = fieldType(**value)
    elif isinstance(value, list):
        innerType = fieldType.__args__[0]
        if innerType in [str, float, bool]:
            attributes[field] = value
        elif innerType is int:
            attributes[field] = [int(x) for x in value]
        elif issubclass(innerType, Enum):
            if len(value) > 0:
                if isinstance(value[0], str):
                    attributes[field] = [innerType(x.upper()) for x in value]
                elif isinstance(value[0], int) and getattr(innerType, "from_proto", None) is not None:
                    attributes[field] = [innerType.from_proto(x) for x in value]
        else:
            attributes[field] = [innerType(**x) for x in value]
    elif isinstance(value, str) and issubclass(fieldType, Enum):
        attributes[field] = fieldType(value.upper())
    elif isinstance(value, int) and issubclass(fieldType, Enum) and getattr(fieldType, "from_proto", None) is not None:
        attributes[field] = fieldType.from_proto(value)
    elif isinstance(value, str) and fieldType == datetime:
        dt = timestamp_pb2.Timestamp()
        dt.FromJsonString(value)
        attributes[field] = dt.ToDatetime()
    elif isinstance(value, str) and fieldType == timedelta:
        td = timestamp_pb2.Duration()
        td.FromJsonString(value)
        attributes[field] = td.ToTimeDelta()
    elif isinstance(value, str) and fieldType == int:
        # Needed to handle int64 encoded as strings.
        attributes[field] = int(value)
    # Set zero values to be consistent with proto.
    elif value is None:
        if fieldType == str:
            attributes[field] = ""
        elif fieldType == int:
            attributes[field] = 0
        elif fieldType == float:
            attributes[field] = 0.0
        elif fieldType == bool:
            attributes[field] = False
        elif getattr(fieldType, "__origin__", None) == list:
            attributes[field] = []
        elif getattr(fieldType, "__origin__", None) == dict:
            attributes[field] = {}
        elif fieldType == datetime or fieldType == timedelta:
            attributes[field] = None
        else:
            if oneOfField == "":
                attributes[field] = fieldType()
            else:
                attributes[field] = None
    elif value is not None and not isinstance(value, fieldType):
        raise ValueError(
            f"'{class_name}.{field}' must have type {fieldType}, got {type(value)}"
        )
    else:
        attributes[field] = value


class Enum(enum.Enum):
    """Base enum class.
    
    Overrides equality to be case insentitive.
    """

    def __eq__(self, other) -> bool:
        """Object is equal with all types."""
        if isinstance(other, str):
            return self.name == other.upper()
        elif isinstance(other, int):
            return self.value == other
        else:
            return self.value == other.value
    
    def __str__(self) -> str:
        return self.value


class GraphQLFieldOptions(Enum):
    """/ Designates a field within a message as being an input- or output-only type."""
    
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    IGNORE = "IGNORE"

    @classmethod
    def from_proto(cls, value: int = 0) ->  GraphQLFieldOptions:
        """Convert enum from proto representation."""
        return cls(options_pb2.GraphQLFieldOptions.Name(value))

    def to_proto(self) -> int:
        """Convert enum to proto representation."""
        return options_pb2.GraphQLFieldOptions.Value(self.value)



from __future__ import annotations
from typing import List, Optional
from google.protobuf import field_mask_pb2
from continual.python.sdk.environments import ProjectEnvironmentManager
from continual.python.utils.utils import (
    generate_datastore_field_paths,
)
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk import client
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.featuresets import FeatureSetManager
from continual.python.sdk.models import ModelManager
from continual.python.sdk.changes import ChangeManager
from continual.python.sdk.events import EventManager
import requests


class OrganizationProjectManager:
    """Manages organization projects."""

    manager: ProjectManager
    parent: str = ""
    client: client.Client

    def __init__(self, parent: str, client: client.Client):
        self.parent = parent
        self.client = client
        self.manager = ProjectManager(parent=self.parent, client=self.client)

    def list(self, page_size: Optional[int] = None) -> List[Project]:
        """List project.

        Arguments:
            page_size: Number of elements.

        Returns:
            A list of organization projects.
        """
        return self.manager.list(page_size)

    def list_all(self) -> Pager[Project]:
        """List all projects.

        Returns:
            An iterator of all organization projects.
        """
        return self.manager.list_all()

    def create(
        self,
        display_name: str,
        data_store: Optional[types.DataStore] = None,
        project_id: Optional[str] = None,
    ) -> Project:
        """Create a new project.

        Creates a new project within organization.  The project ID by
        default is generated from the display name and must be globally
        unique across all organizations.

        Arguments:
            display_name: Display name.
            data_store: Data store configuration.
            project_id: Optional project id.

        Returns:
            A new project
        """
        return self.manager.create(
            display_name=display_name,
            organization=self.parent,
            data_store=data_store,
            project_id=project_id,
        )


class ProjectManager(Manager):
    """Manages project resources."""

    name_pattern: str = "projects/{project}"

    def get(self, id: str) -> Project:
        """Get project.

        Arguments:
            id: Project name or id.

        Returns
            A project.
        """

        project_id = id.split("@")[0]
        req = management_pb2.GetProjectRequest(name=self.name(project_id))
        resp = self.client._management.GetProject(req)
        project = Project.from_proto(resp, client=self.client)
        if "@" in id:
            environment = self.client.environments.get(id.split("@")[1])
            project._set_environment(environment)
        return project

    def list(
        self, page_size: Optional[int] = None, filters: List[str] = None
    ) -> List[Project]:
        """List projects.

        Arguments:
            page_size: Number of items to return.

        Returns:
            A list of projects.
        """
        req = management_pb2.ListProjectsRequest(
            parent=self.parent, page_size=page_size, filters=filters
        )
        resp = self.client._management.ListProjects(req)
        return [Project.from_proto(x, client=self.client) for x in resp.projects]

    def list_all(self) -> Pager[Project]:
        """List all projects.

        Pages through all projects using an iterator.

        Returns:
            A iterator of all projects.
        """

        def next_page(next_page_token):
            req = management_pb2.ListProjectsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListProjects(req)
            return (
                [Project.from_proto(x, client=self.client) for x in resp.projects],
                resp.next_page_token,
            )

        return Pager(next_page)

    def create(
        self,
        display_name: str,
        organization: str,
        data_store: Optional[types.DataStore] = None,
        project_id: Optional[str] = None,
    ) -> Project:
        """Create an project.

        New projects are identified by a unique project id that is
        generated from the display name.  To set a project id explicitly,
        you can pass `project_id`.  However project ids are globally unique across
        all organizations.

        Arguments:
            display_name: Display name.
            organization: Organization resource name.
            data_store: Data store configuration.
            project_id: User-defined project id.

        Returns:
            A new project.
        """
        req = management_pb2.CreateProjectRequest(
            project=Project(
                display_name=display_name,
                organization=organization,
            ).to_proto()
        )
        if data_store is not None:
            req.project.data_store.type = data_store.type
            if data_store.type == "postgres":
                req.project.data_store.postgres.host = data_store.postgres.host
                req.project.data_store.postgres.port = data_store.postgres.port
                req.project.data_store.postgres.username = data_store.postgres.username
                req.project.data_store.postgres.password = data_store.postgres.password
                req.project.data_store.postgres.database = data_store.postgres.database
            elif data_store.type == "snowflake":
                req.project.data_store.snowflake.host = data_store.snowflake.host
                req.project.data_store.snowflake.port = data_store.snowflake.port
                req.project.data_store.snowflake.username = (
                    data_store.snowflake.username
                )
                req.project.data_store.snowflake.password = (
                    data_store.snowflake.password
                )
                req.project.data_store.snowflake.database = (
                    data_store.snowflake.database
                )
                req.project.data_store.snowflake.warehouse = (
                    data_store.snowflake.warehouse
                )
                req.project.data_store.snowflake.role = data_store.snowflake.role

            elif data_store.type == "bigquery":
                req.project.data_store.big_query.auth_file = (
                    data_store.big_query.auth_file
                )
                req.project.data_store.big_query.auth_file_name = (
                    data_store.big_query.auth_file_name
                )
            elif data_store.type == "redshift":
                req.project.data_store.redshift.host = data_store.redshift.host
                req.project.data_store.redshift.port = data_store.redshift.port
                req.project.data_store.redshift.username = data_store.redshift.username
                req.project.data_store.redshift.password = data_store.redshift.password
                req.project.data_store.redshift.database = data_store.redshift.database
            elif data_store.type == "databricks":
                req.project.data_store.databricks.host = data_store.databricks.host
                req.project.data_store.databricks.port = data_store.databricks.port
                req.project.data_store.databricks.token = data_store.databricks.token
                req.project.data_store.databricks.http_path = (
                    data_store.databricks.http_path
                )
                req.project.data_store.databricks.database = (
                    data_store.databricks.database
                )

        resp = self.client._management.CreateProject(req)
        return Project.from_proto(resp, client=self.client)

    def delete(self, id: str, delete_schema: bool = False) -> None:
        """Delete an project.

        Arguments:
            id: Project name or id.
        """

        req = management_pb2.DeleteProjectRequest(
            name=self.name(id), delete_schema=delete_schema
        )
        self.client._management.DeleteProject(req)

    def update(
        self,
        id: str,
        display_name: Optional[str] = None,
        data_store: Optional[types.DataStore] = None,
    ) -> Project:
        """Update project.

        Arguments:
            display_name:  Display name.
            data_store: Project data store.
        Returns:
            Updated project.
        """
        paths = []
        if display_name is not None:
            paths.append("display_name")
        if data_store is not None:
            paths.extend(generate_datastore_field_paths(data_store))

        req = management_pb2.UpdateProjectRequest(
            update_mask=field_mask_pb2.FieldMask(paths=paths),
            project=Project(
                name=self.name(id), display_name=display_name, data_store=data_store
            ).to_proto(),
        )
        resp = self.client._management.UpdateProject(req)
        return Project.from_proto(resp, client=self.client)


class Project(Resource, types.Project):
    """Project resource."""

    name_pattern: str = "projects/{project}"
    manager: ProjectManager

    feature_sets: FeatureSetManager
    """Feature set manager."""

    models: ModelManager
    """Model manager."""

    changes: ChangeManager
    """Change manager."""

    events: EventManager
    """Event manager."""

    environments: ProjectEnvironmentManager

    def _init(self):
        self.manager = ProjectManager(parent=self.parent, client=self.client)
        self.feature_sets = FeatureSetManager(parent=self.name, client=self.client)
        self.models = ModelManager(parent=self.name, client=self.client)
        self.changes = ChangeManager(parent=self.name, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)
        self.environments = ProjectEnvironmentManager(
            parent=self.name, client=self.client
        )
        self.environment = None

    def _set_environment(self, environment):
        self.environment = environment
        self.feature_sets = FeatureSetManager(
            parent=self.environment.name, client=self.client
        )
        self.models = ModelManager(parent=self.environment.name, client=self.client)
        self.changes = ChangeManager(parent=self.environment.name, client=self.client)
        self.events = EventManager(parent=self.environment.name, client=self.client)

    def delete(self, delete_schema: bool = False) -> None:
        """Delete project."""
        self.manager.delete(self.name, delete_schema=delete_schema)

    def update(
        self,
        display_name: Optional[str] = None,
        data_store: Optional[types.DataStore] = None,
    ) -> Project:
        """Update project.

        Arguments:
            display_name:  Display name.

        Returns:
            Updated project.
        """
        return self.manager.update(
            self.name, display_name=display_name, data_store=data_store
        )

    def _create_csv_signed_url(self, file_name: str) -> str:
        """Get a temporary google signed url.

        Arguments:
            file_name: name of the file to get a signed url for.
        """
        name = self.name
        if self.environment is not None:
            name = self.environment.name

        req = management_pb2.CreateFeatureSetCSVSignedURLRequest(
            parent=name, file_name=file_name
        )
        resp = self.client._management.CreateFeatureSetCSVSignedURL(req)
        return resp.signed_url, resp.bucket_path

    def upload_csv_file(self, file_name: str) -> str:
        """upload a csv file to a continual bucket.

        Arguments:
            file_name: name of the file to upload.
        """
        signed_url, bucket_path = self._create_csv_signed_url(file_name)
        TYPE = "text/csv"
        # TODO: chunk up the file if needed
        resp = requests.put(
            signed_url,
            open(file_name).read().encode("utf-8"),
            headers={"Content-Type": TYPE},
        )
        if resp.text != "":
            raise Exception("could not upload file {}: {}".format(file_name, resp.text))

        return bucket_path

    def seed(self, file_name: str, table_name: str = None) -> str:
        """upload a csv file to a continual bucket.

        Arguments:
            file_name: name of the file to upload.
        """
        if not file_name.startswith("gs://"):
            signed_url, bucket_path = self._create_csv_signed_url(file_name)
            TYPE = "text/csv"
            # TODO: chunk up the file if needed
            resp = requests.put(
                signed_url,
                open(file_name, encoding="ISO-8859-1").read().encode("utf-8"),
                headers={"Content-Type": TYPE},
            )
            if resp.text != "":
                raise Exception(
                    "could not upload file {}: {}".format(file_name, resp.text)
                )
        else:
            bucket_path = file_name

        req = management_pb2.SeedRequest(
            project=self.name, file_path=bucket_path, table_name=table_name
        )
        resp = self.client._management.Seed(req)
        return resp.schema_name, resp.table_name

    def get_schema_graph(self) -> str:
        """The a JSON graph of current schema within the project.

        Arguments:
            None
        """

        req = management_pb2.GetSchemaGraphRequest(name=self.name)
        resp = self.client._management.GetSchemaGraph(req)
        return resp.graph_json
